
/* ---------------------------------------------------------------------------

  carray_cast_func.c

  This file is part of Ruby/CArray extension library.
  You can redistribute it and/or modify it under the terms of
  the Ruby's licence.

  This file is automatically generated from carray_math.rb.
  Copyright (C) 2005-2020 Hiroki Motoyoshi

---------------------------------------------------------------------------- */

#include "carray.h"

static void
ca_cast_not_implemented(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
  rb_raise(rb_eCADataTypeError, 
           "can not cast data type from <%s> to <%s>",
           ca_type_name[a1->data_type], ca_type_name[a2->data_type]);
}

/* ------------------ FIXLEN -> FIXLEN ------------------------ */

static void
ca_cast_fixlen_fixlen(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   char *p1 = ptr1;
   ca_size_t bytes1;
   char *p2 = ptr2;
   ca_size_t bytes2;
   if ( a1 == NULL || a2 == NULL ) {
     rb_raise(rb_eRuntimeError, "[BUG] failed to cast fixlen -> fixlen");
   }
   bytes1 = a1->bytes;
   bytes2 = a2->bytes;
   if ( m ) {
     while ( n-- ) {
       if ( !*m ) {
         if ( bytes2 <= bytes1 ) {
           memcpy(p2, p1, bytes2);
         }
         else {
           memset(p2, 0, bytes2);
           memcpy(p2, p1, bytes1);
         }
       }
       p1+=bytes1; p2+=bytes2; m++;
     }
   }
   else {
     while ( n-- ) {
       if ( bytes2 <= bytes1 ) {
         memcpy(p2, p1, bytes2);
       }
       else {
         memset(p2, 0, bytes2);
         memcpy(p2, p1, bytes1);
       }
       p1+=bytes1; p2+=bytes2;
     }
   }
   return;
}

/* ------------------ FIXLEN -> OBJECT ------------------------ */

static void
ca_cast_fixlen_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   char  *p1 = ptr1;
   ca_size_t bytes;
   VALUE *p2 = ptr2;
   if ( a1 == NULL ) {
     rb_raise(rb_eRuntimeError, "[BUG] failed to cast fixlen -> object");
   }
   bytes = a1->bytes;
   if ( m ) {
     while ( n-- ) {
       if ( !*m ) { *p2 = rb_str_new(p1, bytes); OBJ_TAINT(*p2); }
       p1+=bytes; p2++; m++;
     }
   }
   else {
     while ( n-- ) {
       *p2 = rb_str_new(p1, bytes); 
       OBJ_TAINT(*p2); 
       p1+=bytes; p2++;
     }
   }
   return;
}

/* ------------------ BOOLEAN+INTEGER -> BOOLEAN ------------------------ */

static void
ca_cast_boolean8_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   boolean8_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_int8_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   int8_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_uint8_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   uint8_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_int16_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   int16_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_uint16_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   uint16_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_int32_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   int32_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_uint32_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   uint32_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_int64_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   int64_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

static void
ca_cast_uint64_t_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   uint64_t q;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { 
         q = *p1;
         if ( q == 0 || q == 1 ) {
           *p2 = q;
         }
         else {
           rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
         }
       }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { 
       q = *p1;
       if ( q == 0 || q == 1 ) {
         *p2 = q;
       }
       else {
         rb_raise(rb_eRuntimeError, "out of range to cast to boolean (0 or 1)");
       }
       p1++; p2++; 
     }
   }
   return;
}

/* ------------------ BOOLEAN+NUMERIC -> NUMERIC ------------------------ */

static void
ca_cast_boolean8_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_boolean8_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *q1 = ptr1;
   boolean8_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int8_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *q1 = ptr1;
   int8_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint8_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *q1 = ptr1;
   uint8_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int16_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *q1 = ptr1;
   int16_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint16_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *q1 = ptr1;
   uint16_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int32_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *q1 = ptr1;
   int32_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint32_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *q1 = ptr1;
   uint32_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_int64_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *q1 = ptr1;
   int64_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_uint64_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *q1 = ptr1;
   uint64_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float32_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *q1 = ptr1;
   float32_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float64_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *q1 = ptr1;
   float64_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_float128_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *q1 = ptr1;
   float128_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx64_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *q1 = ptr1;
   cmplx64_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx128_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *q1 = ptr1;
   cmplx128_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   int8_t *q2 = ptr2;
   int8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   uint8_t *q2 = ptr2;
   uint8_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint8_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint8_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   int16_t *q2 = ptr2;
   int16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   uint16_t *q2 = ptr2;
   uint16_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint16_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint16_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   int32_t *q2 = ptr2;
   int32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   uint32_t *q2 = ptr2;
   uint32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   int64_t *q2 = ptr2;
   int64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (int64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (int64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   uint64_t *q2 = ptr2;
   uint64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (uint64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (uint64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   float32_t *q2 = ptr2;
   float32_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float32_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float32_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   float64_t *q2 = ptr2;
   float64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   float128_t *q2 = ptr2;
   float128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (float128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (float128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   cmplx64_t *q2 = ptr2;
   cmplx64_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx64_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx64_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   cmplx128_t *q2 = ptr2;
   cmplx128_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx128_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx128_t) *p1;
     }
   }
   return;
}

static void
ca_cast_cmplx256_t_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *q1 = ptr1;
   cmplx256_t *p1 = q1;
   cmplx256_t *q2 = ptr2;
   cmplx256_t *p2 = q2;
   ca_size_t k;
   if ( m ) {
     boolean8_t *pm = m;
     #ifdef _OPENMP
     #pragma omp parallel for private(pm,p1,p2)
     #endif
     for (k=0; k<n; k++) {
       pm = m + k;
       if ( ! *pm ) { 
         p1 = q1 + k;
         p2 = q2 + k;
         *p2 = (cmplx256_t) *p1;
       }
     }
   }
   else {
     #ifdef _OPENMP
     #pragma omp parallel for private(p1,p2)
     #endif
     for (k=0; k<n; k++) {
       p1 = q1 + k;
       p2 = q2 + k;
       *p2 = (cmplx256_t) *p1;
     }
   }
   return;
}

/* ------------------ BOOLEAN+NUMERIC -> OBJECT ------------------------ */

static void
ca_cast_boolean8_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   boolean8_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) BOOL2OBJ(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) BOOL2OBJ(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_int8_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int8_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) LONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) LONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_uint8_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint8_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) ULONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) ULONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_int16_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int16_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) LONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) LONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_uint16_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint16_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) ULONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) ULONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_int32_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int32_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) LONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) LONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_uint32_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint32_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) ULONG2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) ULONG2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_int64_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   int64_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) LL2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) LL2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_uint64_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   uint64_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) ULL2NUM(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) ULL2NUM(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_float32_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float32_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_float_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_float_new(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_float64_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float64_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_float_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_float_new(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_float128_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   float128_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_float_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_float_new(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_cmplx64_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx64_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_ccomplex_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_ccomplex_new(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_cmplx128_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx128_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_ccomplex_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_ccomplex_new(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_cmplx256_t_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   cmplx256_t *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) rb_ccomplex_new(*p1); } 
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) rb_ccomplex_new(*p1); p1++; p2++; }
   }
   return;
}

/* ------------------ OBJECT -> FIXLEN ------------------------ */

static void
ca_cast_VALUE_fixlen(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   char *p2 = ptr2;
   ca_size_t bytes;
   if ( a2 == NULL ) {
     rb_raise(rb_eRuntimeError, "[BUG] failed to cast object -> fixlen");
   }
   bytes = a2->bytes;
   if ( m ) {
     while ( n-- ) {
       if ( !*m ) {
         VALUE str = *p1;
         Check_Type(str, T_STRING);
         if ( bytes <= RSTRING_LEN(str) ) {
           memcpy(p2, StringValuePtr(str), bytes);
         }
         else {
           memset(p2, 0, bytes);
           memcpy(p2, StringValuePtr(str), RSTRING_LEN(str));
         }
       }
       p1++; p2+=bytes; m++;
     }
   }
   else {
   while ( n-- ) {
     VALUE str = *p1;
     Check_Type(str, T_STRING);
     if ( bytes <= RSTRING_LEN(str) ) {
       memcpy(p2, StringValuePtr(str), bytes);
     }
     else {
       memset(p2, 0, bytes);
       memcpy(p2, StringValuePtr(str), RSTRING_LEN(str));
     }
     p1++; p2+=bytes;
   }
   }
   return;
}

/* ------------------ OBJECT -> BOOLEAN+NUMERIC+OBJECT ------------------------ */

static void
ca_cast_VALUE_boolean8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   boolean8_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (boolean8_t) OBJ2BOOL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (boolean8_t) OBJ2BOOL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_int8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   int8_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (int8_t) OBJ2LONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (int8_t) OBJ2LONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_uint8_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   uint8_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (uint8_t) OBJ2ULONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (uint8_t) OBJ2ULONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_int16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   int16_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (int16_t) OBJ2LONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (int16_t) OBJ2LONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_uint16_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   uint16_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (uint16_t) OBJ2ULONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (uint16_t) OBJ2ULONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_int32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   int32_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (int32_t) OBJ2LONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (int32_t) OBJ2LONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_uint32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   uint32_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (uint32_t) OBJ2ULONG(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (uint32_t) OBJ2ULONG(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_int64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   int64_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (int64_t) OBJ2LL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (int64_t) OBJ2LL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_uint64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   uint64_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (uint64_t) OBJ2ULL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (uint64_t) OBJ2ULL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_float32_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   float32_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (float32_t) OBJ2DBL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (float32_t) OBJ2DBL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_float64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   float64_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (float64_t) OBJ2DBL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (float64_t) OBJ2DBL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_float128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   float128_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (float128_t) OBJ2DBL(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (float128_t) OBJ2DBL(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_cmplx64_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   cmplx64_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (cmplx64_t) NUM2CC(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (cmplx64_t) NUM2CC(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_cmplx128_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   cmplx128_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (cmplx128_t) NUM2CC(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (cmplx128_t) NUM2CC(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_cmplx256_t(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   cmplx256_t *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (cmplx256_t) NUM2CC(*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (cmplx256_t) NUM2CC(*p1); p1++; p2++; }
   }
   return;
}

static void
ca_cast_VALUE_VALUE(ca_size_t n, CArray *a1, void *ptr1, CArray *a2, void *ptr2, boolean8_t *m)
{
   VALUE *p1 = ptr1;
   VALUE *p2 = ptr2;
   if ( m ) {
     while ( n-- ) { 
       if ( !*m ) { *p2 = (VALUE) (*p1); }
       p1++; p2++; m++;
     }
   }
   else {
     while ( n-- ) { *p2 = (VALUE) (*p1); p1++; p2++; }
   }
   return;
}

/* ------------------ ca_cast_func_table ------------------------ */

ca_cast_func_t
ca_cast_func_table[CA_NTYPE][CA_NTYPE] = {
  {
    ca_cast_fixlen_fixlen,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_fixlen_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_boolean8_t_boolean8_t,
    ca_cast_boolean8_t_int8_t,
    ca_cast_boolean8_t_uint8_t,
    ca_cast_boolean8_t_int16_t,
    ca_cast_boolean8_t_uint16_t,
    ca_cast_boolean8_t_int32_t,
    ca_cast_boolean8_t_uint32_t,
    ca_cast_boolean8_t_int64_t,
    ca_cast_boolean8_t_uint64_t,
    ca_cast_boolean8_t_float32_t,
    ca_cast_boolean8_t_float64_t,
    ca_cast_boolean8_t_float128_t,
    ca_cast_boolean8_t_cmplx64_t,
    ca_cast_boolean8_t_cmplx128_t,
    ca_cast_boolean8_t_cmplx256_t,
    ca_cast_boolean8_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_int8_t_boolean8_t,
    ca_cast_int8_t_int8_t,
    ca_cast_int8_t_uint8_t,
    ca_cast_int8_t_int16_t,
    ca_cast_int8_t_uint16_t,
    ca_cast_int8_t_int32_t,
    ca_cast_int8_t_uint32_t,
    ca_cast_int8_t_int64_t,
    ca_cast_int8_t_uint64_t,
    ca_cast_int8_t_float32_t,
    ca_cast_int8_t_float64_t,
    ca_cast_int8_t_float128_t,
    ca_cast_int8_t_cmplx64_t,
    ca_cast_int8_t_cmplx128_t,
    ca_cast_int8_t_cmplx256_t,
    ca_cast_int8_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_uint8_t_boolean8_t,
    ca_cast_uint8_t_int8_t,
    ca_cast_uint8_t_uint8_t,
    ca_cast_uint8_t_int16_t,
    ca_cast_uint8_t_uint16_t,
    ca_cast_uint8_t_int32_t,
    ca_cast_uint8_t_uint32_t,
    ca_cast_uint8_t_int64_t,
    ca_cast_uint8_t_uint64_t,
    ca_cast_uint8_t_float32_t,
    ca_cast_uint8_t_float64_t,
    ca_cast_uint8_t_float128_t,
    ca_cast_uint8_t_cmplx64_t,
    ca_cast_uint8_t_cmplx128_t,
    ca_cast_uint8_t_cmplx256_t,
    ca_cast_uint8_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_int16_t_boolean8_t,
    ca_cast_int16_t_int8_t,
    ca_cast_int16_t_uint8_t,
    ca_cast_int16_t_int16_t,
    ca_cast_int16_t_uint16_t,
    ca_cast_int16_t_int32_t,
    ca_cast_int16_t_uint32_t,
    ca_cast_int16_t_int64_t,
    ca_cast_int16_t_uint64_t,
    ca_cast_int16_t_float32_t,
    ca_cast_int16_t_float64_t,
    ca_cast_int16_t_float128_t,
    ca_cast_int16_t_cmplx64_t,
    ca_cast_int16_t_cmplx128_t,
    ca_cast_int16_t_cmplx256_t,
    ca_cast_int16_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_uint16_t_boolean8_t,
    ca_cast_uint16_t_int8_t,
    ca_cast_uint16_t_uint8_t,
    ca_cast_uint16_t_int16_t,
    ca_cast_uint16_t_uint16_t,
    ca_cast_uint16_t_int32_t,
    ca_cast_uint16_t_uint32_t,
    ca_cast_uint16_t_int64_t,
    ca_cast_uint16_t_uint64_t,
    ca_cast_uint16_t_float32_t,
    ca_cast_uint16_t_float64_t,
    ca_cast_uint16_t_float128_t,
    ca_cast_uint16_t_cmplx64_t,
    ca_cast_uint16_t_cmplx128_t,
    ca_cast_uint16_t_cmplx256_t,
    ca_cast_uint16_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_int32_t_boolean8_t,
    ca_cast_int32_t_int8_t,
    ca_cast_int32_t_uint8_t,
    ca_cast_int32_t_int16_t,
    ca_cast_int32_t_uint16_t,
    ca_cast_int32_t_int32_t,
    ca_cast_int32_t_uint32_t,
    ca_cast_int32_t_int64_t,
    ca_cast_int32_t_uint64_t,
    ca_cast_int32_t_float32_t,
    ca_cast_int32_t_float64_t,
    ca_cast_int32_t_float128_t,
    ca_cast_int32_t_cmplx64_t,
    ca_cast_int32_t_cmplx128_t,
    ca_cast_int32_t_cmplx256_t,
    ca_cast_int32_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_uint32_t_boolean8_t,
    ca_cast_uint32_t_int8_t,
    ca_cast_uint32_t_uint8_t,
    ca_cast_uint32_t_int16_t,
    ca_cast_uint32_t_uint16_t,
    ca_cast_uint32_t_int32_t,
    ca_cast_uint32_t_uint32_t,
    ca_cast_uint32_t_int64_t,
    ca_cast_uint32_t_uint64_t,
    ca_cast_uint32_t_float32_t,
    ca_cast_uint32_t_float64_t,
    ca_cast_uint32_t_float128_t,
    ca_cast_uint32_t_cmplx64_t,
    ca_cast_uint32_t_cmplx128_t,
    ca_cast_uint32_t_cmplx256_t,
    ca_cast_uint32_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_int64_t_boolean8_t,
    ca_cast_int64_t_int8_t,
    ca_cast_int64_t_uint8_t,
    ca_cast_int64_t_int16_t,
    ca_cast_int64_t_uint16_t,
    ca_cast_int64_t_int32_t,
    ca_cast_int64_t_uint32_t,
    ca_cast_int64_t_int64_t,
    ca_cast_int64_t_uint64_t,
    ca_cast_int64_t_float32_t,
    ca_cast_int64_t_float64_t,
    ca_cast_int64_t_float128_t,
    ca_cast_int64_t_cmplx64_t,
    ca_cast_int64_t_cmplx128_t,
    ca_cast_int64_t_cmplx256_t,
    ca_cast_int64_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_uint64_t_boolean8_t,
    ca_cast_uint64_t_int8_t,
    ca_cast_uint64_t_uint8_t,
    ca_cast_uint64_t_int16_t,
    ca_cast_uint64_t_uint16_t,
    ca_cast_uint64_t_int32_t,
    ca_cast_uint64_t_uint32_t,
    ca_cast_uint64_t_int64_t,
    ca_cast_uint64_t_uint64_t,
    ca_cast_uint64_t_float32_t,
    ca_cast_uint64_t_float64_t,
    ca_cast_uint64_t_float128_t,
    ca_cast_uint64_t_cmplx64_t,
    ca_cast_uint64_t_cmplx128_t,
    ca_cast_uint64_t_cmplx256_t,
    ca_cast_uint64_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_float32_t_int8_t,
    ca_cast_float32_t_uint8_t,
    ca_cast_float32_t_int16_t,
    ca_cast_float32_t_uint16_t,
    ca_cast_float32_t_int32_t,
    ca_cast_float32_t_uint32_t,
    ca_cast_float32_t_int64_t,
    ca_cast_float32_t_uint64_t,
    ca_cast_float32_t_float32_t,
    ca_cast_float32_t_float64_t,
    ca_cast_float32_t_float128_t,
    ca_cast_float32_t_cmplx64_t,
    ca_cast_float32_t_cmplx128_t,
    ca_cast_float32_t_cmplx256_t,
    ca_cast_float32_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_float64_t_int8_t,
    ca_cast_float64_t_uint8_t,
    ca_cast_float64_t_int16_t,
    ca_cast_float64_t_uint16_t,
    ca_cast_float64_t_int32_t,
    ca_cast_float64_t_uint32_t,
    ca_cast_float64_t_int64_t,
    ca_cast_float64_t_uint64_t,
    ca_cast_float64_t_float32_t,
    ca_cast_float64_t_float64_t,
    ca_cast_float64_t_float128_t,
    ca_cast_float64_t_cmplx64_t,
    ca_cast_float64_t_cmplx128_t,
    ca_cast_float64_t_cmplx256_t,
    ca_cast_float64_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_float128_t_int8_t,
    ca_cast_float128_t_uint8_t,
    ca_cast_float128_t_int16_t,
    ca_cast_float128_t_uint16_t,
    ca_cast_float128_t_int32_t,
    ca_cast_float128_t_uint32_t,
    ca_cast_float128_t_int64_t,
    ca_cast_float128_t_uint64_t,
    ca_cast_float128_t_float32_t,
    ca_cast_float128_t_float64_t,
    ca_cast_float128_t_float128_t,
    ca_cast_float128_t_cmplx64_t,
    ca_cast_float128_t_cmplx128_t,
    ca_cast_float128_t_cmplx256_t,
    ca_cast_float128_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_cmplx64_t_int8_t,
    ca_cast_cmplx64_t_uint8_t,
    ca_cast_cmplx64_t_int16_t,
    ca_cast_cmplx64_t_uint16_t,
    ca_cast_cmplx64_t_int32_t,
    ca_cast_cmplx64_t_uint32_t,
    ca_cast_cmplx64_t_int64_t,
    ca_cast_cmplx64_t_uint64_t,
    ca_cast_cmplx64_t_float32_t,
    ca_cast_cmplx64_t_float64_t,
    ca_cast_cmplx64_t_float128_t,
    ca_cast_cmplx64_t_cmplx64_t,
    ca_cast_cmplx64_t_cmplx128_t,
    ca_cast_cmplx64_t_cmplx256_t,
    ca_cast_cmplx64_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_cmplx128_t_int8_t,
    ca_cast_cmplx128_t_uint8_t,
    ca_cast_cmplx128_t_int16_t,
    ca_cast_cmplx128_t_uint16_t,
    ca_cast_cmplx128_t_int32_t,
    ca_cast_cmplx128_t_uint32_t,
    ca_cast_cmplx128_t_int64_t,
    ca_cast_cmplx128_t_uint64_t,
    ca_cast_cmplx128_t_float32_t,
    ca_cast_cmplx128_t_float64_t,
    ca_cast_cmplx128_t_float128_t,
    ca_cast_cmplx128_t_cmplx64_t,
    ca_cast_cmplx128_t_cmplx128_t,
    ca_cast_cmplx128_t_cmplx256_t,
    ca_cast_cmplx128_t_VALUE
  },
  {
    ca_cast_not_implemented,
    ca_cast_not_implemented,
    ca_cast_cmplx256_t_int8_t,
    ca_cast_cmplx256_t_uint8_t,
    ca_cast_cmplx256_t_int16_t,
    ca_cast_cmplx256_t_uint16_t,
    ca_cast_cmplx256_t_int32_t,
    ca_cast_cmplx256_t_uint32_t,
    ca_cast_cmplx256_t_int64_t,
    ca_cast_cmplx256_t_uint64_t,
    ca_cast_cmplx256_t_float32_t,
    ca_cast_cmplx256_t_float64_t,
    ca_cast_cmplx256_t_float128_t,
    ca_cast_cmplx256_t_cmplx64_t,
    ca_cast_cmplx256_t_cmplx128_t,
    ca_cast_cmplx256_t_cmplx256_t,
    ca_cast_cmplx256_t_VALUE
  },
  {
    ca_cast_VALUE_fixlen,
    ca_cast_VALUE_boolean8_t,
    ca_cast_VALUE_int8_t,
    ca_cast_VALUE_uint8_t,
    ca_cast_VALUE_int16_t,
    ca_cast_VALUE_uint16_t,
    ca_cast_VALUE_int32_t,
    ca_cast_VALUE_uint32_t,
    ca_cast_VALUE_int64_t,
    ca_cast_VALUE_uint64_t,
    ca_cast_VALUE_float32_t,
    ca_cast_VALUE_float64_t,
    ca_cast_VALUE_float128_t,
    ca_cast_VALUE_cmplx64_t,
    ca_cast_VALUE_cmplx128_t,
    ca_cast_VALUE_cmplx256_t,
    ca_cast_VALUE_VALUE
  }
};

