/* ---------------------------------------------------------------------------

  carray_math.c

  This file is part of Ruby/CArray extension library.


  This file is automatically generated from carray_math.rb.
  Copyright (C) 2005-2020 Hiroki Motoyoshi

---------------------------------------------------------------------------- */

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#endif

#include "ruby.h"

static ID id_equal, id_lt, id_le, id_gt, id_ge;
static ID id_uminus, id_utilda;
static ID id_plus, id_minus, id_star, id_slash, id_percent, id_star_star;
static ID id_and, id_or, id_xor, id_lshift, id_rshift;



#include <math.h>
#include "carray.h"


/*----------------------- zero --------------------------*/

static void
ca_monop_zero_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0; 
      }
    }
  }
}

static void
ca_monop_zero_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0.0; 
      }
    }
  }
}

static void
ca_monop_zero_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0.0; 
      }
    }
  }
}

static void
ca_monop_zero_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0.0; 
      }
    }
  }
}

static void
ca_monop_zero_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

ca_monop_func_t
ca_monop_zero[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_zero_boolean8_t,
  ca_monop_zero_int8_t,
  ca_monop_zero_uint8_t,
  ca_monop_zero_int16_t,
  ca_monop_zero_uint16_t,
  ca_monop_zero_int32_t,
  ca_monop_zero_uint32_t,
  ca_monop_zero_int64_t,
  ca_monop_zero_uint64_t,
  ca_monop_zero_float32_t,
  ca_monop_zero_float64_t,
  ca_monop_zero_float128_t,
  ca_monop_zero_cmplx64_t,
  ca_monop_zero_cmplx128_t,
  ca_monop_zero_cmplx256_t,
  ca_monop_zero_VALUE,
};


static VALUE rb_ca_zero (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_zero); }
static VALUE rb_ca_zero_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_zero); }
  
/*----------------------- one --------------------------*/

static void
ca_monop_one_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1; 
      }
    }
  }
}

static void
ca_monop_one_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1.0; 
      }
    }
  }
}

static void
ca_monop_one_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1.0; 
      }
    }
  }
}

static void
ca_monop_one_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1.0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1.0; 
      }
    }
  }
}

static void
ca_monop_one_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 3; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 3; 
      }
    }
  }
}

ca_monop_func_t
ca_monop_one[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_one_boolean8_t,
  ca_monop_one_int8_t,
  ca_monop_one_uint8_t,
  ca_monop_one_int16_t,
  ca_monop_one_uint16_t,
  ca_monop_one_int32_t,
  ca_monop_one_uint32_t,
  ca_monop_one_int64_t,
  ca_monop_one_uint64_t,
  ca_monop_one_float32_t,
  ca_monop_one_float64_t,
  ca_monop_one_float128_t,
  ca_monop_one_cmplx64_t,
  ca_monop_one_cmplx128_t,
  ca_monop_one_cmplx256_t,
  ca_monop_one_VALUE,
};


static VALUE rb_ca_one (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_one); }
static VALUE rb_ca_one_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_one); }
  
/*----------------------- frac --------------------------*/

static void
ca_monop_frac_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 0;  
      }
    }
  }
}

static void
ca_monop_frac_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
      }
    }
  }
}

static void
ca_monop_frac_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
      }
    }
  }
}

static void
ca_monop_frac_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? (*p1)-floor(*p1) : ((*p1)<0.0) ? (*p1)-ceil(*p1) : 0.0;  
      }
    }
  }
}

static void
ca_monop_frac_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("frac"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("frac"), 0); 
      }
    }
  }
}

ca_monop_func_t
ca_monop_frac[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_frac_int8_t,
  ca_monop_frac_uint8_t,
  ca_monop_frac_int16_t,
  ca_monop_frac_uint16_t,
  ca_monop_frac_int32_t,
  ca_monop_frac_uint32_t,
  ca_monop_frac_int64_t,
  ca_monop_frac_uint64_t,
  ca_monop_frac_float32_t,
  ca_monop_frac_float64_t,
  ca_monop_frac_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_frac_VALUE,
};


static VALUE rb_ca_frac (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_frac); }
static VALUE rb_ca_frac_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_frac); }
  
/*----------------------- neg --------------------------*/

static void
ca_monop_neg_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = -(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = -(*p1); 
      }
    }
  }
}

static void
ca_monop_neg_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), id_uminus, 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), id_uminus, 0); 
      }
    }
  }
}

ca_monop_func_t
ca_monop_neg[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_neg_int8_t,
  ca_monop_neg_uint8_t,
  ca_monop_neg_int16_t,
  ca_monop_neg_uint16_t,
  ca_monop_neg_int32_t,
  ca_monop_neg_uint32_t,
  ca_monop_neg_int64_t,
  ca_monop_neg_uint64_t,
  ca_monop_neg_float32_t,
  ca_monop_neg_float64_t,
  ca_monop_neg_float128_t,
  ca_monop_neg_cmplx64_t,
  ca_monop_neg_cmplx128_t,
  ca_monop_neg_cmplx256_t,
  ca_monop_neg_VALUE,
};


static VALUE rb_ca_neg (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_neg); }
static VALUE rb_ca_neg_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_neg); }
  
/*----------------------- bit_neg --------------------------*/

static void
ca_monop_bit_neg_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1) ? 0 : 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1) ? 0 : 1; 
      }
    }
  }
}

static void
ca_monop_bit_neg_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ~(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ~(*p1); 
      }
    }
  }
}

static void
ca_monop_bit_neg_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), id_utilda, 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), id_utilda, 0); 
      }
    }
  }
}

ca_monop_func_t
ca_monop_bit_neg[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_bit_neg_boolean8_t,
  ca_monop_bit_neg_int8_t,
  ca_monop_bit_neg_uint8_t,
  ca_monop_bit_neg_int16_t,
  ca_monop_bit_neg_uint16_t,
  ca_monop_bit_neg_int32_t,
  ca_monop_bit_neg_uint32_t,
  ca_monop_bit_neg_int64_t,
  ca_monop_bit_neg_uint64_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_bit_neg_VALUE,
};


static VALUE rb_ca_bit_neg (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_bit_neg); }
static VALUE rb_ca_bit_neg_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_bit_neg); }
  
/*----------------------- abs_i --------------------------*/

static void
ca_monop_abs_i_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = abs(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = abs(*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = fabs((float64_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = fabs((float64_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = fabs((float64_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = fabs((float64_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = fabs((float64_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = fabs((float64_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cabs((cmplx128_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cabs((cmplx128_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cabs((cmplx128_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cabs((cmplx128_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cabs((cmplx128_t)*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cabs((cmplx128_t)*p1); 
      }
    }
  }
}

static void
ca_monop_abs_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("abs"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("abs"), 0); 
      }
    }
  }
}

ca_monop_func_t
ca_monop_abs_i[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_abs_i_int8_t,
  ca_monop_abs_i_uint8_t,
  ca_monop_abs_i_int16_t,
  ca_monop_abs_i_uint16_t,
  ca_monop_abs_i_int32_t,
  ca_monop_abs_i_uint32_t,
  ca_monop_abs_i_int64_t,
  ca_monop_abs_i_uint64_t,
  ca_monop_abs_i_float32_t,
  ca_monop_abs_i_float64_t,
  ca_monop_abs_i_float128_t,
  ca_monop_abs_i_cmplx64_t,
  ca_monop_abs_i_cmplx128_t,
  ca_monop_abs_i_cmplx256_t,
  ca_monop_abs_i_VALUE,
};


static VALUE rb_ca_abs_i (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_abs_i); }
static VALUE rb_ca_abs_i_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_abs_i); }
  

static VALUE
rb_ca_abs (VALUE self)
{
  CArray *ca;

  Data_Get_Struct(self, CArray, ca);
  if ( ca_is_complex_type(ca) ) {
    VALUE ret = rb_ca_abs_i(self);
    return rb_ca_copy(rb_funcall(ret, rb_intern("real"), 0));
  }
  else {
    return rb_ca_abs_i(self);
  }
}

static VALUE
rb_ca_abs_bang (VALUE self)
{
  CArray *ca;
  rb_ca_modify(self);
  Data_Get_Struct(self, CArray, ca);
  if ( ca_is_complex_type(ca) ) {
    VALUE ret = rb_ca_abs_i(self);
    return rb_funcall(self, rb_intern("[]="), 1, ret);
  }
  else {
    return rb_ca_abs_i_bang(self);
  }
}

static VALUE rb_cmath_abs (VALUE mod, VALUE arg)
{ return rb_ca_abs(arg); }


/*----------------------- conj --------------------------*/

static void
ca_monop_conj_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1); 
      }
    }
  }
}

static void
ca_monop_conj_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = conj(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = conj(*p1); 
      }
    }
  }
}

static void
ca_monop_conj_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = conj(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = conj(*p1); 
      }
    }
  }
}

static void
ca_monop_conj_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = conj(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = conj(*p1); 
      }
    }
  }
}

static void
ca_monop_conj_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("conj"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("conj"), 0); 
      }
    }
  }
}

ca_monop_func_t
ca_monop_conj[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_conj_int8_t,
  ca_monop_conj_uint8_t,
  ca_monop_conj_int16_t,
  ca_monop_conj_uint16_t,
  ca_monop_conj_int32_t,
  ca_monop_conj_uint32_t,
  ca_monop_conj_int64_t,
  ca_monop_conj_uint64_t,
  ca_monop_conj_float32_t,
  ca_monop_conj_float64_t,
  ca_monop_conj_float128_t,
  ca_monop_conj_cmplx64_t,
  ca_monop_conj_cmplx128_t,
  ca_monop_conj_cmplx256_t,
  ca_monop_conj_VALUE,
};


static VALUE rb_ca_conj (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_conj); }
static VALUE rb_ca_conj_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_conj); }
  
/*----------------------- rad --------------------------*/

static void
ca_monop_rad_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (0.0174532925199433*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (0.0174532925199433*(*p1));
      }
    }
  }
}

static void
ca_monop_rad_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (0.0174532925199433*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (0.0174532925199433*(*p1));
      }
    }
  }
}

static void
ca_monop_rad_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (0.0174532925199433*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (0.0174532925199433*(*p1));
      }
    }
  }
}

static void
ca_monop_rad_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("rad"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("rad"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_rad[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_rad_float32_t,
  ca_monop_rad_float64_t,
  ca_monop_rad_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_rad_VALUE,
};


static VALUE rb_ca_rad (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_rad); 
}

static VALUE rb_ca_rad_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_rad); }

static VALUE rb_cmath_rad (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("rad")); }

/*----------------------- deg --------------------------*/

static void
ca_monop_deg_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (57.2957795130823*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (57.2957795130823*(*p1));
      }
    }
  }
}

static void
ca_monop_deg_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (57.2957795130823*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (57.2957795130823*(*p1));
      }
    }
  }
}

static void
ca_monop_deg_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (57.2957795130823*(*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (57.2957795130823*(*p1));
      }
    }
  }
}

static void
ca_monop_deg_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("deg"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("deg"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_deg[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_deg_float32_t,
  ca_monop_deg_float64_t,
  ca_monop_deg_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_deg_VALUE,
};


static VALUE rb_ca_deg (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_deg); 
}

static VALUE rb_ca_deg_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_deg); }

static VALUE rb_cmath_deg (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("deg")); }

/*----------------------- ceil --------------------------*/

static void
ca_monop_ceil_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_ceil_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ceil(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ceil(*p1);
      }
    }
  }
}

static void
ca_monop_ceil_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ceil(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ceil(*p1);
      }
    }
  }
}

static void
ca_monop_ceil_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ceil(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ceil(*p1);
      }
    }
  }
}

static void
ca_monop_ceil_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("ceil"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("ceil"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_ceil[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_ceil_int8_t,
  ca_monop_ceil_uint8_t,
  ca_monop_ceil_int16_t,
  ca_monop_ceil_uint16_t,
  ca_monop_ceil_int32_t,
  ca_monop_ceil_uint32_t,
  ca_monop_ceil_int64_t,
  ca_monop_ceil_uint64_t,
  ca_monop_ceil_float32_t,
  ca_monop_ceil_float64_t,
  ca_monop_ceil_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_ceil_VALUE,
};


static VALUE rb_ca_ceil (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_ceil); }

static VALUE rb_ca_ceil_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_ceil); }

static VALUE rb_cmath_ceil (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("ceil")); }

/*----------------------- floor --------------------------*/

static void
ca_monop_floor_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_floor_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = floor(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = floor(*p1);
      }
    }
  }
}

static void
ca_monop_floor_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = floor(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = floor(*p1);
      }
    }
  }
}

static void
ca_monop_floor_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = floor(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = floor(*p1);
      }
    }
  }
}

static void
ca_monop_floor_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("floor"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("floor"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_floor[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_floor_int8_t,
  ca_monop_floor_uint8_t,
  ca_monop_floor_int16_t,
  ca_monop_floor_uint16_t,
  ca_monop_floor_int32_t,
  ca_monop_floor_uint32_t,
  ca_monop_floor_int64_t,
  ca_monop_floor_uint64_t,
  ca_monop_floor_float32_t,
  ca_monop_floor_float64_t,
  ca_monop_floor_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_floor_VALUE,
};


static VALUE rb_ca_floor (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_floor); }

static VALUE rb_ca_floor_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_floor); }

static VALUE rb_cmath_floor (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("floor")); }

/*----------------------- round --------------------------*/

static void
ca_monop_round_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1);
      }
    }
  }
}

static void
ca_monop_round_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0; 
      }
    }
  }
}

static void
ca_monop_round_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0; 
      }
    }
  }
}

static void
ca_monop_round_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0;  
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ((*p1)>0.0) ? floor((*p1)+0.5) : ((*p1)<0.0) ? ceil((*p1)-0.5) : 0.0; 
      }
    }
  }
}

static void
ca_monop_round_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("round"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("round"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_round[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_round_int8_t,
  ca_monop_round_uint8_t,
  ca_monop_round_int16_t,
  ca_monop_round_uint16_t,
  ca_monop_round_int32_t,
  ca_monop_round_uint32_t,
  ca_monop_round_int64_t,
  ca_monop_round_uint64_t,
  ca_monop_round_float32_t,
  ca_monop_round_float64_t,
  ca_monop_round_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_round_VALUE,
};


static VALUE rb_ca_round (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_round); }

static VALUE rb_ca_round_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_round); }

static VALUE rb_cmath_round (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("round")); }

/*----------------------- rcp --------------------------*/

static void
ca_monop_rcp_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1/(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = 1/(*p1);
      }
    }
  }
}

static void
ca_monop_rcp_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall(INT2NUM(1), id_slash, 1, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall(INT2NUM(1), id_slash, 1, (*p1));
      }
    }
  }
}

ca_monop_func_t
ca_monop_rcp[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_rcp_int8_t,
  ca_monop_rcp_uint8_t,
  ca_monop_rcp_int16_t,
  ca_monop_rcp_uint16_t,
  ca_monop_rcp_int32_t,
  ca_monop_rcp_uint32_t,
  ca_monop_rcp_int64_t,
  ca_monop_rcp_uint64_t,
  ca_monop_rcp_float32_t,
  ca_monop_rcp_float64_t,
  ca_monop_rcp_float128_t,
  ca_monop_rcp_cmplx64_t,
  ca_monop_rcp_cmplx128_t,
  ca_monop_rcp_cmplx256_t,
  ca_monop_rcp_VALUE,
};


static VALUE rb_ca_rcp (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_rcp); }

static VALUE rb_ca_rcp_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_rcp); }

static VALUE rb_cmath_rcp (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("rcp")); }

/*----------------------- sqrt --------------------------*/

static void
ca_monop_sqrt_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csqrt(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csqrt(*p1);
      }
    }
  }
}

static void
ca_monop_sqrt_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("sqrt"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("sqrt"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_sqrt[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_sqrt_float32_t,
  ca_monop_sqrt_float64_t,
  ca_monop_sqrt_float128_t,
  ca_monop_sqrt_cmplx64_t,
  ca_monop_sqrt_cmplx128_t,
  ca_monop_sqrt_cmplx256_t,
  ca_monop_sqrt_VALUE,
};


static VALUE rb_ca_sqrt (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_sqrt); 
}

static VALUE rb_ca_sqrt_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_sqrt); }

static VALUE rb_cmath_sqrt (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("sqrt")); }

/*----------------------- exp --------------------------*/

static void
ca_monop_exp_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cexp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cexp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cexp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cexp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cexp(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cexp(*p1);
      }
    }
  }
}

static void
ca_monop_exp_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("exp"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("exp"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_exp[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_exp_float32_t,
  ca_monop_exp_float64_t,
  ca_monop_exp_float128_t,
  ca_monop_exp_cmplx64_t,
  ca_monop_exp_cmplx128_t,
  ca_monop_exp_cmplx256_t,
  ca_monop_exp_VALUE,
};


static VALUE rb_ca_exp (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_exp); 
}

static VALUE rb_ca_exp_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_exp); }

static VALUE rb_cmath_exp (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("exp")); }

/*----------------------- exp2 --------------------------*/

static void
ca_monop_exp2_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp2(*p1);
      }
    }
  }
}

static void
ca_monop_exp2_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp2(*p1);
      }
    }
  }
}

static void
ca_monop_exp2_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = exp2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = exp2(*p1);
      }
    }
  }
}

static void
ca_monop_exp2_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(2, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(2, (*p1));
      }
    }
  }
}

static void
ca_monop_exp2_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(2, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(2, (*p1));
      }
    }
  }
}

static void
ca_monop_exp2_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(2, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(2, (*p1));
      }
    }
  }
}

static void
ca_monop_exp2_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("exp2"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("exp2"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_exp2[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_exp2_float32_t,
  ca_monop_exp2_float64_t,
  ca_monop_exp2_float128_t,
  ca_monop_exp2_cmplx64_t,
  ca_monop_exp2_cmplx128_t,
  ca_monop_exp2_cmplx256_t,
  ca_monop_exp2_VALUE,
};


static VALUE rb_ca_exp2 (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_exp2); 
}

static VALUE rb_ca_exp2_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_exp2); }

static VALUE rb_cmath_exp2 (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("exp2")); }

/*----------------------- exp10 --------------------------*/

static void
ca_monop_exp10_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = pow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = pow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = pow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = pow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = pow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = pow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cpow(10, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cpow(10, (*p1));
      }
    }
  }
}

static void
ca_monop_exp10_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall(INT2NUM(10), id_star_star, 1, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall(INT2NUM(10), id_star_star, 1, (*p1));
      }
    }
  }
}

ca_monop_func_t
ca_monop_exp10[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_exp10_float32_t,
  ca_monop_exp10_float64_t,
  ca_monop_exp10_float128_t,
  ca_monop_exp10_cmplx64_t,
  ca_monop_exp10_cmplx128_t,
  ca_monop_exp10_cmplx256_t,
  ca_monop_exp10_VALUE,
};


static VALUE rb_ca_exp10 (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_exp10); 
}

static VALUE rb_ca_exp10_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_exp10); }

static VALUE rb_cmath_exp10 (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("exp10")); }

/*----------------------- log --------------------------*/

static void
ca_monop_log_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log(*p1);
      }
    }
  }
}

static void
ca_monop_log_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log(*p1);
      }
    }
  }
}

static void
ca_monop_log_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log(*p1);
      }
    }
  }
}

static void
ca_monop_log_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = clog(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = clog(*p1);
      }
    }
  }
}

static void
ca_monop_log_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = clog(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = clog(*p1);
      }
    }
  }
}

static void
ca_monop_log_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = clog(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = clog(*p1);
      }
    }
  }
}

static void
ca_monop_log_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("log"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("log"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_log[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_log_float32_t,
  ca_monop_log_float64_t,
  ca_monop_log_float128_t,
  ca_monop_log_cmplx64_t,
  ca_monop_log_cmplx128_t,
  ca_monop_log_cmplx256_t,
  ca_monop_log_VALUE,
};


static VALUE rb_ca_log (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_log); 
}

static VALUE rb_ca_log_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_log); }

static VALUE rb_cmath_log (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("log")); }

/*----------------------- log10 --------------------------*/

static void
ca_monop_log10_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log10(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log10(*p1);
      }
    }
  }
}

static void
ca_monop_log10_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log10(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log10(*p1);
      }
    }
  }
}

static void
ca_monop_log10_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log10(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log10(*p1);
      }
    }
  }
}

static void
ca_monop_log10_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("log10"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("log10"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_log10[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_log10_float32_t,
  ca_monop_log10_float64_t,
  ca_monop_log10_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_log10_VALUE,
};


static VALUE rb_ca_log10 (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_log10); 
}

static VALUE rb_ca_log10_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_log10); }

static VALUE rb_cmath_log10 (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("log10")); }

/*----------------------- log2 --------------------------*/

static void
ca_monop_log2_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log2(*p1);
      }
    }
  }
}

static void
ca_monop_log2_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log2(*p1);
      }
    }
  }
}

static void
ca_monop_log2_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = log2(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = log2(*p1);
      }
    }
  }
}

static void
ca_monop_log2_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("log2"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("log2"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_log2[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_log2_float32_t,
  ca_monop_log2_float64_t,
  ca_monop_log2_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_log2_VALUE,
};


static VALUE rb_ca_log2 (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_log2); 
}

static VALUE rb_ca_log2_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_log2); }

static VALUE rb_cmath_log2 (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("log2")); }

/*----------------------- logb --------------------------*/

static void
ca_monop_logb_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = logb(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = logb(*p1);
      }
    }
  }
}

static void
ca_monop_logb_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = logb(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = logb(*p1);
      }
    }
  }
}

static void
ca_monop_logb_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = logb(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = logb(*p1);
      }
    }
  }
}

static void
ca_monop_logb_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("logb"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("logb"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_logb[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_logb_float32_t,
  ca_monop_logb_float64_t,
  ca_monop_logb_float128_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_logb_VALUE,
};


static VALUE rb_ca_logb (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_logb); 
}

static VALUE rb_ca_logb_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_logb); }

static VALUE rb_cmath_logb (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("logb")); }

/*----------------------- sin --------------------------*/

static void
ca_monop_sin_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = csin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = csin(*p1);
      }
    }
  }
}

static void
ca_monop_sin_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("sin"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("sin"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_sin[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_sin_float32_t,
  ca_monop_sin_float64_t,
  ca_monop_sin_float128_t,
  ca_monop_sin_cmplx64_t,
  ca_monop_sin_cmplx128_t,
  ca_monop_sin_cmplx256_t,
  ca_monop_sin_VALUE,
};


static VALUE rb_ca_sin (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_sin); 
}

static VALUE rb_ca_sin_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_sin); }

static VALUE rb_cmath_sin (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("sin")); }

/*----------------------- cos --------------------------*/

static void
ca_monop_cos_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ccos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ccos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ccos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ccos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ccos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ccos(*p1);
      }
    }
  }
}

static void
ca_monop_cos_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("cos"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("cos"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_cos[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_cos_float32_t,
  ca_monop_cos_float64_t,
  ca_monop_cos_float128_t,
  ca_monop_cos_cmplx64_t,
  ca_monop_cos_cmplx128_t,
  ca_monop_cos_cmplx256_t,
  ca_monop_cos_VALUE,
};


static VALUE rb_ca_cos (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_cos); 
}

static VALUE rb_ca_cos_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_cos); }

static VALUE rb_cmath_cos (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("cos")); }

/*----------------------- tan --------------------------*/

static void
ca_monop_tan_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ctan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ctan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ctan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ctan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = ctan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = ctan(*p1);
      }
    }
  }
}

static void
ca_monop_tan_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("tan"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("tan"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_tan[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_tan_float32_t,
  ca_monop_tan_float64_t,
  ca_monop_tan_float128_t,
  ca_monop_tan_cmplx64_t,
  ca_monop_tan_cmplx128_t,
  ca_monop_tan_cmplx256_t,
  ca_monop_tan_VALUE,
};


static VALUE rb_ca_tan (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_tan); 
}

static VALUE rb_ca_tan_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_tan); }

static VALUE rb_cmath_tan (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("tan")); }

/*----------------------- asin --------------------------*/

static void
ca_monop_asin_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = casin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = casin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = casin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = casin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = casin(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = casin(*p1);
      }
    }
  }
}

static void
ca_monop_asin_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("asin"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("asin"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_asin[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_asin_float32_t,
  ca_monop_asin_float64_t,
  ca_monop_asin_float128_t,
  ca_monop_asin_cmplx64_t,
  ca_monop_asin_cmplx128_t,
  ca_monop_asin_cmplx256_t,
  ca_monop_asin_VALUE,
};


static VALUE rb_ca_asin (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_asin); 
}

static VALUE rb_ca_asin_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_asin); }

static VALUE rb_cmath_asin (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("asin")); }

/*----------------------- acos --------------------------*/

static void
ca_monop_acos_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cacos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cacos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cacos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cacos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cacos(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cacos(*p1);
      }
    }
  }
}

static void
ca_monop_acos_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("acos"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("acos"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_acos[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_acos_float32_t,
  ca_monop_acos_float64_t,
  ca_monop_acos_float128_t,
  ca_monop_acos_cmplx64_t,
  ca_monop_acos_cmplx128_t,
  ca_monop_acos_cmplx256_t,
  ca_monop_acos_VALUE,
};


static VALUE rb_ca_acos (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_acos); 
}

static VALUE rb_ca_acos_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_acos); }

static VALUE rb_cmath_acos (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("acos")); }

/*----------------------- atan --------------------------*/

static void
ca_monop_atan_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = catan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = catan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = catan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = catan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = catan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = catan(*p1);
      }
    }
  }
}

static void
ca_monop_atan_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("atan"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("atan"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_atan[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_atan_float32_t,
  ca_monop_atan_float64_t,
  ca_monop_atan_float128_t,
  ca_monop_atan_cmplx64_t,
  ca_monop_atan_cmplx128_t,
  ca_monop_atan_cmplx256_t,
  ca_monop_atan_VALUE,
};


static VALUE rb_ca_atan (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_atan); 
}

static VALUE rb_ca_atan_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_atan); }

static VALUE rb_cmath_atan (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("atan")); }

/*----------------------- sinh --------------------------*/

static void
ca_monop_sinh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = sinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = sinh(*p1);
      }
    }
  }
}

static void
ca_monop_sinh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("sinh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("sinh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_sinh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_sinh_float32_t,
  ca_monop_sinh_float64_t,
  ca_monop_sinh_float128_t,
  ca_monop_sinh_cmplx64_t,
  ca_monop_sinh_cmplx128_t,
  ca_monop_sinh_cmplx256_t,
  ca_monop_sinh_VALUE,
};


static VALUE rb_ca_sinh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_sinh); 
}

static VALUE rb_ca_sinh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_sinh); }

static VALUE rb_cmath_sinh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("sinh")); }

/*----------------------- cosh --------------------------*/

static void
ca_monop_cosh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = cosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = cosh(*p1);
      }
    }
  }
}

static void
ca_monop_cosh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("cosh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("cosh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_cosh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_cosh_float32_t,
  ca_monop_cosh_float64_t,
  ca_monop_cosh_float128_t,
  ca_monop_cosh_cmplx64_t,
  ca_monop_cosh_cmplx128_t,
  ca_monop_cosh_cmplx256_t,
  ca_monop_cosh_VALUE,
};


static VALUE rb_ca_cosh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_cosh); 
}

static VALUE rb_ca_cosh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_cosh); }

static VALUE rb_cmath_cosh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("cosh")); }

/*----------------------- tanh --------------------------*/

static void
ca_monop_tanh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = tanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = tanh(*p1);
      }
    }
  }
}

static void
ca_monop_tanh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("tanh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("tanh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_tanh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_tanh_float32_t,
  ca_monop_tanh_float64_t,
  ca_monop_tanh_float128_t,
  ca_monop_tanh_cmplx64_t,
  ca_monop_tanh_cmplx128_t,
  ca_monop_tanh_cmplx256_t,
  ca_monop_tanh_VALUE,
};


static VALUE rb_ca_tanh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_tanh); 
}

static VALUE rb_ca_tanh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_tanh); }

static VALUE rb_cmath_tanh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("tanh")); }

/*----------------------- asinh --------------------------*/

static void
ca_monop_asinh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = asinh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = asinh(*p1);
      }
    }
  }
}

static void
ca_monop_asinh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("asinh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("asinh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_asinh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_asinh_float32_t,
  ca_monop_asinh_float64_t,
  ca_monop_asinh_float128_t,
  ca_monop_asinh_cmplx64_t,
  ca_monop_asinh_cmplx128_t,
  ca_monop_asinh_cmplx256_t,
  ca_monop_asinh_VALUE,
};


static VALUE rb_ca_asinh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_asinh); 
}

static VALUE rb_ca_asinh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_asinh); }

static VALUE rb_cmath_asinh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("asinh")); }

/*----------------------- acosh --------------------------*/

static void
ca_monop_acosh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = acosh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = acosh(*p1);
      }
    }
  }
}

static void
ca_monop_acosh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("acosh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("acosh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_acosh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_acosh_float32_t,
  ca_monop_acosh_float64_t,
  ca_monop_acosh_float128_t,
  ca_monop_acosh_cmplx64_t,
  ca_monop_acosh_cmplx128_t,
  ca_monop_acosh_cmplx256_t,
  ca_monop_acosh_VALUE,
};


static VALUE rb_ca_acosh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_acosh); 
}

static VALUE rb_ca_acosh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_acosh); }

static VALUE rb_cmath_acosh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("acosh")); }

/*----------------------- atanh --------------------------*/

static void
ca_monop_atanh_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = atanh(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = atanh(*p1);
      }
    }
  }
}

static void
ca_monop_atanh_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("atanh"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("atanh"), 0);
      }
    }
  }
}

ca_monop_func_t
ca_monop_atanh[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_atanh_float32_t,
  ca_monop_atanh_float64_t,
  ca_monop_atanh_float128_t,
  ca_monop_atanh_cmplx64_t,
  ca_monop_atanh_cmplx128_t,
  ca_monop_atanh_cmplx256_t,
  ca_monop_atanh_VALUE,
};


static VALUE rb_ca_atanh (VALUE self)
{ 
  if ( rb_ca_is_integer_type(self) ) {
    self = rb_ca_wrap_readonly(self, INT2NUM(CA_FLOAT64));
  }
  return rb_ca_call_monop(self, ca_monop_atanh); 
}

static VALUE rb_ca_atanh_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_atanh); }

static VALUE rb_cmath_atanh (VALUE mod, VALUE arg)
{ return ca_math_call(mod, arg, rb_intern("atanh")); }

/*----------------------- pmax --------------------------*/

static void
ca_binop_pmax_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) > (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) > (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmax_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmax(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmax(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmax_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmax(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmax(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmax_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmax(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmax(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmax_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall(rb_assoc_new((*p1),(*p2)), rb_intern("max"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall(rb_assoc_new((*p1),(*p2)), rb_intern("max"), 0);
      }
    }
  }
}

ca_binop_func_t
ca_binop_pmax[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_pmax_int8_t,
  ca_binop_pmax_uint8_t,
  ca_binop_pmax_int16_t,
  ca_binop_pmax_uint16_t,
  ca_binop_pmax_int32_t,
  ca_binop_pmax_uint32_t,
  ca_binop_pmax_int64_t,
  ca_binop_pmax_uint64_t,
  ca_binop_pmax_float32_t,
  ca_binop_pmax_float64_t,
  ca_binop_pmax_float128_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_pmax_VALUE,
};


static VALUE rb_ca_pmax (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("pmax"));
  }
  return rb_ca_call_binop(self, other, ca_binop_pmax); 
}
static VALUE rb_ca_pmax_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_pmax); }
  
/*----------------------- pmin --------------------------*/

static void
ca_binop_pmin_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) < (*p2) ? (*p1) : (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) < (*p2) ? (*p1) : (*p2);
      }
    }
  }
}

static void
ca_binop_pmin_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmin(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmin(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmin_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmin(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmin(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmin_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmin(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmin(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_pmin_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall(rb_assoc_new((*p1),(*p2)), rb_intern("min"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall(rb_assoc_new((*p1),(*p2)), rb_intern("min"), 0);
      }
    }
  }
}

ca_binop_func_t
ca_binop_pmin[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_pmin_int8_t,
  ca_binop_pmin_uint8_t,
  ca_binop_pmin_int16_t,
  ca_binop_pmin_uint16_t,
  ca_binop_pmin_int32_t,
  ca_binop_pmin_uint32_t,
  ca_binop_pmin_int64_t,
  ca_binop_pmin_uint64_t,
  ca_binop_pmin_float32_t,
  ca_binop_pmin_float64_t,
  ca_binop_pmin_float128_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_pmin_VALUE,
};


static VALUE rb_ca_pmin (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("pmin"));
  }
  return rb_ca_call_binop(self, other, ca_binop_pmin); 
}
static VALUE rb_ca_pmin_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_pmin); }
  
/*----------------------- add --------------------------*/

static void
ca_binop_add_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) + (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) + (*p2);
      }
    }
  }
}

static void
ca_binop_add_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_plus, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_plus, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_add[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_add_int8_t,
  ca_binop_add_uint8_t,
  ca_binop_add_int16_t,
  ca_binop_add_uint16_t,
  ca_binop_add_int32_t,
  ca_binop_add_uint32_t,
  ca_binop_add_int64_t,
  ca_binop_add_uint64_t,
  ca_binop_add_float32_t,
  ca_binop_add_float64_t,
  ca_binop_add_float128_t,
  ca_binop_add_cmplx64_t,
  ca_binop_add_cmplx128_t,
  ca_binop_add_cmplx256_t,
  ca_binop_add_VALUE,
};


static VALUE rb_ca_add (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("+"));
  }
  return rb_ca_call_binop(self, other, ca_binop_add); 
}
static VALUE rb_ca_add_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_add); }
  
/*----------------------- sub --------------------------*/

static void
ca_binop_sub_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) - (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) - (*p2);
      }
    }
  }
}

static void
ca_binop_sub_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_minus, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_minus, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_sub[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_sub_int8_t,
  ca_binop_sub_uint8_t,
  ca_binop_sub_int16_t,
  ca_binop_sub_uint16_t,
  ca_binop_sub_int32_t,
  ca_binop_sub_uint32_t,
  ca_binop_sub_int64_t,
  ca_binop_sub_uint64_t,
  ca_binop_sub_float32_t,
  ca_binop_sub_float64_t,
  ca_binop_sub_float128_t,
  ca_binop_sub_cmplx64_t,
  ca_binop_sub_cmplx128_t,
  ca_binop_sub_cmplx256_t,
  ca_binop_sub_VALUE,
};


static VALUE rb_ca_sub (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("-"));
  }
  return rb_ca_call_binop(self, other, ca_binop_sub); 
}
static VALUE rb_ca_sub_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_sub); }
  
/*----------------------- mul --------------------------*/

static void
ca_binop_mul_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) * (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) * (*p2);
      }
    }
  }
}

static void
ca_binop_mul_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_star, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_star, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_mul[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_mul_int8_t,
  ca_binop_mul_uint8_t,
  ca_binop_mul_int16_t,
  ca_binop_mul_uint16_t,
  ca_binop_mul_int32_t,
  ca_binop_mul_uint32_t,
  ca_binop_mul_int64_t,
  ca_binop_mul_uint64_t,
  ca_binop_mul_float32_t,
  ca_binop_mul_float64_t,
  ca_binop_mul_float128_t,
  ca_binop_mul_cmplx64_t,
  ca_binop_mul_cmplx128_t,
  ca_binop_mul_cmplx256_t,
  ca_binop_mul_VALUE,
};


static VALUE rb_ca_mul (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("*"));
  }
  return rb_ca_call_binop(self, other, ca_binop_mul); 
}
static VALUE rb_ca_mul_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_mul); }
  
/*----------------------- div --------------------------*/

static void
ca_binop_div_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) / (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) / (*p2);
      }
    }
  }
}

static void
ca_binop_div_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_slash, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_slash, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_div[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_div_int8_t,
  ca_binop_div_uint8_t,
  ca_binop_div_int16_t,
  ca_binop_div_uint16_t,
  ca_binop_div_int32_t,
  ca_binop_div_uint32_t,
  ca_binop_div_int64_t,
  ca_binop_div_uint64_t,
  ca_binop_div_float32_t,
  ca_binop_div_float64_t,
  ca_binop_div_float128_t,
  ca_binop_div_cmplx64_t,
  ca_binop_div_cmplx128_t,
  ca_binop_div_cmplx256_t,
  ca_binop_div_VALUE,
};


static VALUE rb_ca_div (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("/"));
  }
  return rb_ca_call_binop(self, other, ca_binop_div); 
}
static VALUE rb_ca_div_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_div); }
  
/*----------------------- quo_i --------------------------*/

static void
ca_binop_quo_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), rb_intern("quo"), 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), rb_intern("quo"), 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_quo_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_quo_i_VALUE,
};


static VALUE rb_ca_quo_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("quo_i"));
  }
  return rb_ca_call_binop(self, other, ca_binop_quo_i); 
}
static VALUE rb_ca_quo_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_quo_i); }
  
/*----------------------- rcp_mul --------------------------*/

static void
ca_binop_rcp_mul_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p1)==0) {ca_zerodiv();}; (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p2) / (*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p2) / (*p1);
      }
    }
  }
}

static void
ca_binop_rcp_mul_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p2), id_slash, 1, (*p1)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p2), id_slash, 1, (*p1));
      }
    }
  }
}

ca_binop_func_t
ca_binop_rcp_mul[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_rcp_mul_int8_t,
  ca_binop_rcp_mul_uint8_t,
  ca_binop_rcp_mul_int16_t,
  ca_binop_rcp_mul_uint16_t,
  ca_binop_rcp_mul_int32_t,
  ca_binop_rcp_mul_uint32_t,
  ca_binop_rcp_mul_int64_t,
  ca_binop_rcp_mul_uint64_t,
  ca_binop_rcp_mul_float32_t,
  ca_binop_rcp_mul_float64_t,
  ca_binop_rcp_mul_float128_t,
  ca_binop_rcp_mul_cmplx64_t,
  ca_binop_rcp_mul_cmplx128_t,
  ca_binop_rcp_mul_cmplx256_t,
  ca_binop_rcp_mul_VALUE,
};


static VALUE rb_ca_rcp_mul (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("rcp_mul"));
  }
  return rb_ca_call_binop(self, other, ca_binop_rcp_mul); 
}
static VALUE rb_ca_rcp_mul_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_rcp_mul); }
  
/*----------------------- mod --------------------------*/

static void
ca_binop_mod_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_mod_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmod(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmod(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_mod_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmod(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmod(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_mod_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = fmod(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = fmod(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_mod_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_percent, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_percent, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_mod[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_mod_int8_t,
  ca_binop_mod_uint8_t,
  ca_binop_mod_int16_t,
  ca_binop_mod_uint16_t,
  ca_binop_mod_int32_t,
  ca_binop_mod_uint32_t,
  ca_binop_mod_int64_t,
  ca_binop_mod_uint64_t,
  ca_binop_mod_float32_t,
  ca_binop_mod_float64_t,
  ca_binop_mod_float128_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_mod_VALUE,
};


static VALUE rb_ca_mod (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("%"));
  }
  return rb_ca_call_binop(self, other, ca_binop_mod); 
}
static VALUE rb_ca_mod_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_mod); }
  
/*----------------------- reminder --------------------------*/

static void
ca_binop_reminder_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        if ((*p2)==0) {ca_zerodiv();}; (*p3) = (*p1) % (*p2);
      }
    }
  }
}

static void
ca_binop_reminder_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = remainder(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = remainder(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_reminder_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = remainder(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = remainder(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_reminder_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = remainder(*p1, *p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = remainder(*p1, *p2);
      }
    }
  }
}

static void
ca_binop_reminder_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_percent, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_percent, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_reminder[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_reminder_int8_t,
  ca_binop_reminder_uint8_t,
  ca_binop_reminder_int16_t,
  ca_binop_reminder_uint16_t,
  ca_binop_reminder_int32_t,
  ca_binop_reminder_uint32_t,
  ca_binop_reminder_int64_t,
  ca_binop_reminder_uint64_t,
  ca_binop_reminder_float32_t,
  ca_binop_reminder_float64_t,
  ca_binop_reminder_float128_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_reminder_VALUE,
};


static VALUE rb_ca_reminder (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("reminder"));
  }
  return rb_ca_call_binop(self, other, ca_binop_reminder); 
}
static VALUE rb_ca_reminder_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_reminder); }
  
/*----------------------- bit_and_i --------------------------*/

static void
ca_binop_bit_and_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) & (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) & (*p2);
      }
    }
  }
}

static void
ca_binop_bit_and_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_and, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_and, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_bit_and_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_bit_and_i_boolean8_t,
  ca_binop_bit_and_i_int8_t,
  ca_binop_bit_and_i_uint8_t,
  ca_binop_bit_and_i_int16_t,
  ca_binop_bit_and_i_uint16_t,
  ca_binop_bit_and_i_int32_t,
  ca_binop_bit_and_i_uint32_t,
  ca_binop_bit_and_i_int64_t,
  ca_binop_bit_and_i_uint64_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_and_i_VALUE,
};


static VALUE rb_ca_bit_and_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("&"));
  }
  return rb_ca_call_binop(self, other, ca_binop_bit_and_i); 
}
static VALUE rb_ca_bit_and_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_bit_and_i); }
  
/*----------------------- bit_or_i --------------------------*/

static void
ca_binop_bit_or_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) | (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) | (*p2);
      }
    }
  }
}

static void
ca_binop_bit_or_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_or, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_or, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_bit_or_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_bit_or_i_boolean8_t,
  ca_binop_bit_or_i_int8_t,
  ca_binop_bit_or_i_uint8_t,
  ca_binop_bit_or_i_int16_t,
  ca_binop_bit_or_i_uint16_t,
  ca_binop_bit_or_i_int32_t,
  ca_binop_bit_or_i_uint32_t,
  ca_binop_bit_or_i_int64_t,
  ca_binop_bit_or_i_uint64_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_or_i_VALUE,
};


static VALUE rb_ca_bit_or_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("|"));
  }
  return rb_ca_call_binop(self, other, ca_binop_bit_or_i); 
}
static VALUE rb_ca_bit_or_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_bit_or_i); }
  
/*----------------------- bit_xor_i --------------------------*/

static void
ca_binop_bit_xor_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ((*p1) != (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = ((*p1) != (*p2)) ? 1 : 0;
      }
    }
  }
}

static void
ca_binop_bit_xor_i_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) ^ (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) ^ (*p2);
      }
    }
  }
}

static void
ca_binop_bit_xor_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_xor, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_xor, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_bit_xor_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_bit_xor_i_boolean8_t,
  ca_binop_bit_xor_i_int8_t,
  ca_binop_bit_xor_i_uint8_t,
  ca_binop_bit_xor_i_int16_t,
  ca_binop_bit_xor_i_uint16_t,
  ca_binop_bit_xor_i_int32_t,
  ca_binop_bit_xor_i_uint32_t,
  ca_binop_bit_xor_i_int64_t,
  ca_binop_bit_xor_i_uint64_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_xor_i_VALUE,
};


static VALUE rb_ca_bit_xor_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("^"));
  }
  return rb_ca_call_binop(self, other, ca_binop_bit_xor_i); 
}
static VALUE rb_ca_bit_xor_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_bit_xor_i); }
  

static VALUE
rb_ca_bit_and (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_bit_and_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_bit_and_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_bit_and_i(self, other);
  }
}

static VALUE
rb_ca_bit_or (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_bit_or_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_bit_or_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_bit_or_i(self, other);
  }
}

static VALUE
rb_ca_bit_xor (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_bit_xor_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_bit_xor_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_bit_xor_i(self, other);
  }
}


/*----------------------- bit_lshift --------------------------*/

static void
ca_binop_bit_lshift_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) << (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) << (*p2);
      }
    }
  }
}

static void
ca_binop_bit_lshift_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_lshift, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_lshift, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_bit_lshift[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_lshift_int8_t,
  ca_binop_bit_lshift_uint8_t,
  ca_binop_bit_lshift_int16_t,
  ca_binop_bit_lshift_uint16_t,
  ca_binop_bit_lshift_int32_t,
  ca_binop_bit_lshift_uint32_t,
  ca_binop_bit_lshift_int64_t,
  ca_binop_bit_lshift_uint64_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_lshift_VALUE,
};


static VALUE rb_ca_bit_lshift (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("<<"));
  }
  return rb_ca_call_binop(self, other, ca_binop_bit_lshift); 
}
static VALUE rb_ca_bit_lshift_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_bit_lshift); }
  
/*----------------------- bit_rshift --------------------------*/

static void
ca_binop_bit_rshift_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) >> (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) >> (*p2);
      }
    }
  }
}

static void
ca_binop_bit_rshift_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_rshift, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_rshift, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_bit_rshift[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_rshift_int8_t,
  ca_binop_bit_rshift_uint8_t,
  ca_binop_bit_rshift_int16_t,
  ca_binop_bit_rshift_uint16_t,
  ca_binop_bit_rshift_int32_t,
  ca_binop_bit_rshift_uint32_t,
  ca_binop_bit_rshift_int64_t,
  ca_binop_bit_rshift_uint64_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_bit_rshift_VALUE,
};


static VALUE rb_ca_bit_rshift (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern(">>"));
  }
  return rb_ca_call_binop(self, other, ca_binop_bit_rshift); 
}
static VALUE rb_ca_bit_rshift_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_bit_rshift); }
  
/*----------------------- feq --------------------------*/

static void
ca_bincmp_feq_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          
         float32_t f1a = fabs((float64_t) *p1);
         float32_t f2a = fabs((float64_t) *p2);
         float32_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * FLT_EPSILON ) ? 1 : 0;
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        
         float32_t f1a = fabs((float64_t) *p1);
         float32_t f2a = fabs((float64_t) *p2);
         float32_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * FLT_EPSILON ) ? 1 : 0;
       
      }
    }
  }
}

static void
ca_bincmp_feq_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          
         float64_t f1a = fabs((float64_t) *p1);
         float64_t f2a = fabs((float64_t) *p2);
         float64_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * DBL_EPSILON ) ? 1 : 0;
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        
         float64_t f1a = fabs((float64_t) *p1);
         float64_t f2a = fabs((float64_t) *p2);
         float64_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * DBL_EPSILON ) ? 1 : 0;
       
      }
    }
  }
}

static void
ca_bincmp_feq_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          
         float128_t f1a = fabs((float64_t) *p1);
         float128_t f2a = fabs((float64_t) *p2);
         float128_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * DBL_EPSILON ) ? 1 : 0;
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        
         float128_t f1a = fabs((float64_t) *p1);
         float128_t f2a = fabs((float64_t) *p2);
         float128_t fmax = (f1a > f2a) ? f1a : f2a;
         (*p3) = ( fabs(((float64_t) *p1)-( (float64_t) *p2)) <= fmax * DBL_EPSILON ) ? 1 : 0;
       
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_feq[CA_NTYPE] = {
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_feq_float32_t,
  ca_bincmp_feq_float64_t,
  ca_bincmp_feq_float128_t,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
};


static VALUE rb_ca_feq (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("feq"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_feq);
}
  
/*----------------------- eq --------------------------*/

static void
ca_bincmp_eq_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          (*p3) = ( b1 == b2 && (! memcmp(p1, p2, b1)) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        (*p3) = ( b1 == b2 && (! memcmp(p1, p2, b1)) );
      }
    }    
  }
}

static void
ca_bincmp_eq_boolean8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) == (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) == (*p2) );
      }
    }
  }
}

static void
ca_bincmp_eq_cmplx64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) == (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) == (*p2);
      }
    }
  }
}

static void
ca_bincmp_eq_cmplx128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) == (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) == (*p2);
      }
    }
  }
}

static void
ca_bincmp_eq_cmplx256_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) == (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) == (*p2);
      }
    }
  }
}

static void
ca_bincmp_eq_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_equal, 1, (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_equal, 1, (*p2)) ? 1 : 0;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_eq[CA_NTYPE] = {
  ca_bincmp_eq_fixlen,
  ca_bincmp_eq_boolean8_t,
  ca_bincmp_eq_int8_t,
  ca_bincmp_eq_uint8_t,
  ca_bincmp_eq_int16_t,
  ca_bincmp_eq_uint16_t,
  ca_bincmp_eq_int32_t,
  ca_bincmp_eq_uint32_t,
  ca_bincmp_eq_int64_t,
  ca_bincmp_eq_uint64_t,
  ca_bincmp_eq_float32_t,
  ca_bincmp_eq_float64_t,
  ca_bincmp_eq_float128_t,
  ca_bincmp_eq_cmplx64_t,
  ca_bincmp_eq_cmplx128_t,
  ca_bincmp_eq_cmplx256_t,
  ca_bincmp_eq_VALUE,
};


static VALUE rb_ca_eq (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("eq"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_eq);
}
  
/*----------------------- ne --------------------------*/

static void
ca_bincmp_ne_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          (*p3) = ( b1 != b2 || memcmp(p1, p2, b1) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        (*p3) = ( b1 != b2 || memcmp(p1, p2, b1) );
      }
    }    
  }
}

static void
ca_bincmp_ne_boolean8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) != (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) != (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ne_cmplx64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2;
  cmplx64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) != (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) != (*p2);
      }
    }
  }
}

static void
ca_bincmp_ne_cmplx128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2;
  cmplx128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) != (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) != (*p2);
      }
    }
  }
}

static void
ca_bincmp_ne_cmplx256_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2;
  cmplx256_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) != (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = (*p1) != (*p2);
      }
    }
  }
}

static void
ca_bincmp_ne_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_equal, 1, (*p2)) ? 0 : 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_equal, 1, (*p2)) ? 0 : 1;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_ne[CA_NTYPE] = {
  ca_bincmp_ne_fixlen,
  ca_bincmp_ne_boolean8_t,
  ca_bincmp_ne_int8_t,
  ca_bincmp_ne_uint8_t,
  ca_bincmp_ne_int16_t,
  ca_bincmp_ne_uint16_t,
  ca_bincmp_ne_int32_t,
  ca_bincmp_ne_uint32_t,
  ca_bincmp_ne_int64_t,
  ca_bincmp_ne_uint64_t,
  ca_bincmp_ne_float32_t,
  ca_bincmp_ne_float64_t,
  ca_bincmp_ne_float128_t,
  ca_bincmp_ne_cmplx64_t,
  ca_bincmp_ne_cmplx128_t,
  ca_bincmp_ne_cmplx256_t,
  ca_bincmp_ne_VALUE,
};


static VALUE rb_ca_ne (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("ne"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_ne);
}
  
/*----------------------- gt --------------------------*/

static void
ca_bincmp_gt_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp > 0 || ( cmp == 0 && b1 > b2 ) );
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp > 0 || ( cmp == 0 && b1 > b2 ) );
       
      }
    }    
  }
}

static void
ca_bincmp_gt_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) > (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) > (*p2) );
      }
    }
  }
}

static void
ca_bincmp_gt_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_gt, 1, (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_gt, 1, (*p2)) ? 1 : 0;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_gt[CA_NTYPE] = {
  ca_bincmp_gt_fixlen,
  ca_bincmp_not_implement,
  ca_bincmp_gt_int8_t,
  ca_bincmp_gt_uint8_t,
  ca_bincmp_gt_int16_t,
  ca_bincmp_gt_uint16_t,
  ca_bincmp_gt_int32_t,
  ca_bincmp_gt_uint32_t,
  ca_bincmp_gt_int64_t,
  ca_bincmp_gt_uint64_t,
  ca_bincmp_gt_float32_t,
  ca_bincmp_gt_float64_t,
  ca_bincmp_gt_float128_t,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_gt_VALUE,
};


static VALUE rb_ca_gt (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("gt"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_gt);
}
  
/*----------------------- lt --------------------------*/

static void
ca_bincmp_lt_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp < 0 || ( cmp == 0 && b1 < b2 ) );
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp < 0 || ( cmp == 0 && b1 < b2 ) );
       
      }
    }    
  }
}

static void
ca_bincmp_lt_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) < (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) < (*p2) );
      }
    }
  }
}

static void
ca_bincmp_lt_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_lt, 1, (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_lt, 1, (*p2)) ? 1 : 0;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_lt[CA_NTYPE] = {
  ca_bincmp_lt_fixlen,
  ca_bincmp_not_implement,
  ca_bincmp_lt_int8_t,
  ca_bincmp_lt_uint8_t,
  ca_bincmp_lt_int16_t,
  ca_bincmp_lt_uint16_t,
  ca_bincmp_lt_int32_t,
  ca_bincmp_lt_uint32_t,
  ca_bincmp_lt_int64_t,
  ca_bincmp_lt_uint64_t,
  ca_bincmp_lt_float32_t,
  ca_bincmp_lt_float64_t,
  ca_bincmp_lt_float128_t,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_lt_VALUE,
};


static VALUE rb_ca_lt (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("lt"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_lt);
}
  
/*----------------------- ge --------------------------*/

static void
ca_bincmp_ge_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp > 0 || ( cmp == 0 && b1 >= b2 ) );
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp > 0 || ( cmp == 0 && b1 >= b2 ) );
       
      }
    }    
  }
}

static void
ca_bincmp_ge_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) >= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) >= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_ge_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_ge, 1, (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_ge, 1, (*p2)) ? 1 : 0;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_ge[CA_NTYPE] = {
  ca_bincmp_ge_fixlen,
  ca_bincmp_not_implement,
  ca_bincmp_ge_int8_t,
  ca_bincmp_ge_uint8_t,
  ca_bincmp_ge_int16_t,
  ca_bincmp_ge_uint16_t,
  ca_bincmp_ge_int32_t,
  ca_bincmp_ge_uint32_t,
  ca_bincmp_ge_int64_t,
  ca_bincmp_ge_uint64_t,
  ca_bincmp_ge_float32_t,
  ca_bincmp_ge_float64_t,
  ca_bincmp_ge_float128_t,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_ge_VALUE,
};


static VALUE rb_ca_ge (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("ge"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_ge);
}
  
/*----------------------- le --------------------------*/

static void
ca_bincmp_le_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp < 0 || ( cmp == 0 && b1 <= b2 ) );
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        
         int cmp = memcmp(p1, p2, b1 < b2 ? b1 : b2) ;
         (*p3) = ( cmp < 0 || ( cmp == 0 && b1 <= b2 ) );
       
      }
    }    
  }
}

static void
ca_bincmp_le_int8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2;
  int8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_uint8_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2;
  uint8_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_int16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2;
  int16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_uint16_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2;
  uint16_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_int32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2;
  int32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_uint32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2;
  uint32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_int64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2;
  int64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_uint64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2;
  uint64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_float32_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2;
  float32_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_float64_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2;
  float64_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_float128_t (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2;
  float128_t *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ( (*p1) <= (*p2) ); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = ( (*p1) <= (*p2) );
      }
    }
  }
}

static void
ca_bincmp_le_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_le, 1, (*p2)) ? 1 : 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = rb_funcall((*p1), id_le, 1, (*p2)) ? 1 : 0;
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_le[CA_NTYPE] = {
  ca_bincmp_le_fixlen,
  ca_bincmp_not_implement,
  ca_bincmp_le_int8_t,
  ca_bincmp_le_uint8_t,
  ca_bincmp_le_int16_t,
  ca_bincmp_le_uint16_t,
  ca_bincmp_le_int32_t,
  ca_bincmp_le_uint32_t,
  ca_bincmp_le_int64_t,
  ca_bincmp_le_uint64_t,
  ca_bincmp_le_float32_t,
  ca_bincmp_le_float64_t,
  ca_bincmp_le_float128_t,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_le_VALUE,
};


static VALUE rb_ca_le (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("le"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_le);
}
  
/*----------------------- not --------------------------*/

static void
ca_monop_not_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2;
  boolean8_t *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (*p1) ? 0 : 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (*p1) ? 0 : 1; 
      }
    }
  }
}

static void
ca_monop_not_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = (RTEST(*p1)) ? Qfalse : Qtrue; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      {
        (*p2) = (RTEST(*p1)) ? Qfalse : Qtrue; 
      }
    }
  }
}

ca_monop_func_t
ca_monop_not[CA_NTYPE] = {
  ca_monop_not_implement,
  ca_monop_not_boolean8_t,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_implement,
  ca_monop_not_VALUE,
};


static VALUE rb_ca_not (VALUE self)
{ return rb_ca_call_monop(self, ca_monop_not); }
static VALUE rb_ca_not_bang (VALUE self)
{ return rb_ca_call_monop_bang(self, ca_monop_not); }
  
/*----------------------- and_i --------------------------*/

static void
ca_binop_and_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) && (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) && (*p2);
      }
    }
  }
}

static void
ca_binop_and_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ((RTEST(*p1)!=0) && (RTEST(*p2)!=0)) ? Qtrue : Qfalse; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = ((RTEST(*p1)!=0) && (RTEST(*p2)!=0)) ? Qtrue : Qfalse;
      }
    }
  }
}

ca_binop_func_t
ca_binop_and_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_and_i_boolean8_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_and_i_VALUE,
};


static VALUE rb_ca_and_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("and"));
  }
  return rb_ca_call_binop(self, other, ca_binop_and_i); 
}
static VALUE rb_ca_and_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_and_i); }
  
/*----------------------- or_i --------------------------*/

static void
ca_binop_or_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (*p1) || (*p2); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (*p1) || (*p2);
      }
    }
  }
}

static void
ca_binop_or_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ((RTEST(*p1)!=0) || (RTEST(*p2)!=0)) ? Qtrue : Qfalse; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = ((RTEST(*p1)!=0) || (RTEST(*p2)!=0)) ? Qtrue : Qfalse;
      }
    }
  }
}

ca_binop_func_t
ca_binop_or_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_or_i_boolean8_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_or_i_VALUE,
};


static VALUE rb_ca_or_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("or"));
  }
  return rb_ca_call_binop(self, other, ca_binop_or_i); 
}
static VALUE rb_ca_or_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_or_i); }
  
/*----------------------- xor_i --------------------------*/

static void
ca_binop_xor_i_boolean8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  boolean8_t *q1 = (boolean8_t *) ptr1, *q2 = (boolean8_t *) ptr2, *q3 = (boolean8_t *) ptr3;
  boolean8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = (((*p1)==0) == ((*p2)==0)) ? 0 : 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = (((*p1)==0) == ((*p2)==0)) ? 0 : 1;
      }
    }
  }
}

static void
ca_binop_xor_i_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = ((RTEST(*p1)) == (RTEST(*p2))) ? Qfalse : Qtrue; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = ((RTEST(*p1)) == (RTEST(*p2))) ? Qfalse : Qtrue;
      }
    }
  }
}

ca_binop_func_t
ca_binop_xor_i[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_xor_i_boolean8_t,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_xor_i_VALUE,
};


static VALUE rb_ca_xor_i (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("xor"));
  }
  return rb_ca_call_binop(self, other, ca_binop_xor_i); 
}
static VALUE rb_ca_xor_i_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_xor_i); }
  

static VALUE
rb_ca_and (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_and_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_and_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_and_i(self, other);
  }
}

static VALUE
rb_ca_or (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_or_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_or_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_or_i(self, other);
  }
}

static VALUE
rb_ca_xor (VALUE self, VALUE other)
{
  if ( rb_ca_is_boolean_type(self) ) {
    return rb_ca_xor_i(self, rb_ca_wrap_readonly(other, INT2NUM(CA_BOOLEAN)));
  }
  else if ( rb_obj_is_carray(other) && rb_ca_is_boolean_type(other) ) {
    return rb_ca_xor_i(rb_ca_wrap_readonly(self, INT2NUM(CA_BOOLEAN)), other);
  }
  else {
    return rb_ca_xor_i(self, other);
  }
}


/*----------------------- is_nan --------------------------*/

static void
ca_moncmp_is_nan_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1;
  int8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1;
  uint8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1;
  int16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1;
  uint16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1;
  int32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1;
  uint32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1;
  int64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1;
  uint64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_nan_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1;
  float32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isnan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isnan(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_nan_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1;
  float64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isnan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isnan(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_nan_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1;
  float128_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isnan(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isnan(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_nan_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1;
  VALUE *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("nan?"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("nan?"), 0);
      }
    }
  }
}

ca_moncmp_func_t
ca_moncmp_is_nan[CA_NTYPE] = {
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_nan_int8_t,
  ca_moncmp_is_nan_uint8_t,
  ca_moncmp_is_nan_int16_t,
  ca_moncmp_is_nan_uint16_t,
  ca_moncmp_is_nan_int32_t,
  ca_moncmp_is_nan_uint32_t,
  ca_moncmp_is_nan_int64_t,
  ca_moncmp_is_nan_uint64_t,
  ca_moncmp_is_nan_float32_t,
  ca_moncmp_is_nan_float64_t,
  ca_moncmp_is_nan_float128_t,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_nan_VALUE,
};


static VALUE rb_ca_is_nan (VALUE self)
{ return rb_ca_call_moncmp(self, ca_moncmp_is_nan); }
  
/*----------------------- is_inf --------------------------*/

static void
ca_moncmp_is_inf_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1;
  int8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1;
  uint8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1;
  int16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1;
  uint16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1;
  int32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1;
  uint32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1;
  int64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1;
  uint64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 0; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 0;
      }
    }
  }
}

static void
ca_moncmp_is_inf_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1;
  float32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isinf(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isinf(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_inf_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1;
  float64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isinf(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isinf(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_inf_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1;
  float128_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isinf(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isinf(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_inf_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1;
  VALUE *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("infinite?"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("infinite?"), 0);
      }
    }
  }
}

ca_moncmp_func_t
ca_moncmp_is_inf[CA_NTYPE] = {
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_inf_int8_t,
  ca_moncmp_is_inf_uint8_t,
  ca_moncmp_is_inf_int16_t,
  ca_moncmp_is_inf_uint16_t,
  ca_moncmp_is_inf_int32_t,
  ca_moncmp_is_inf_uint32_t,
  ca_moncmp_is_inf_int64_t,
  ca_moncmp_is_inf_uint64_t,
  ca_moncmp_is_inf_float32_t,
  ca_moncmp_is_inf_float64_t,
  ca_moncmp_is_inf_float128_t,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_inf_VALUE,
};


static VALUE rb_ca_is_inf (VALUE self)
{ return rb_ca_call_moncmp(self, ca_moncmp_is_inf); }
  
/*----------------------- is_finite --------------------------*/

static void
ca_moncmp_is_finite_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int8_t *q1 = (int8_t *) ptr1;
  int8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint8_t *q1 = (uint8_t *) ptr1;
  uint8_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int16_t *q1 = (int16_t *) ptr1;
  int16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint16_t *q1 = (uint16_t *) ptr1;
  uint16_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int32_t *q1 = (int32_t *) ptr1;
  int32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint32_t *q1 = (uint32_t *) ptr1;
  uint32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  int64_t *q1 = (int64_t *) ptr1;
  int64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  uint64_t *q1 = (uint64_t *) ptr1;
  uint64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = 1; 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = 1;
      }
    }
  }
}

static void
ca_moncmp_is_finite_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float32_t *q1 = (float32_t *) ptr1;
  float32_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isfinite(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isfinite(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_finite_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float64_t *q1 = (float64_t *) ptr1;
  float64_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isfinite(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isfinite(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_finite_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  float128_t *q1 = (float128_t *) ptr1;
  float128_t *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = isfinite(*p1); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = isfinite(*p1);
      }
    }
  }
}

static void
ca_moncmp_is_finite_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, boolean8_t *ptr2, ca_size_t i2)
{
  VALUE *q1 = (VALUE *) ptr1;
  VALUE *p1 = q1;
  boolean8_t *q2 = (boolean8_t *) ptr2;
  boolean8_t *p2 = q2;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        {
          (*p2) = rb_funcall((*p1), rb_intern("finite?"), 0); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      {
        (*p2) = rb_funcall((*p1), rb_intern("finite?"), 0);
      }
    }
  }
}

ca_moncmp_func_t
ca_moncmp_is_finite[CA_NTYPE] = {
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_finite_int8_t,
  ca_moncmp_is_finite_uint8_t,
  ca_moncmp_is_finite_int16_t,
  ca_moncmp_is_finite_uint16_t,
  ca_moncmp_is_finite_int32_t,
  ca_moncmp_is_finite_uint32_t,
  ca_moncmp_is_finite_int64_t,
  ca_moncmp_is_finite_uint64_t,
  ca_moncmp_is_finite_float32_t,
  ca_moncmp_is_finite_float64_t,
  ca_moncmp_is_finite_float128_t,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_not_implement,
  ca_moncmp_is_finite_VALUE,
};


static VALUE rb_ca_is_finite (VALUE self)
{ return rb_ca_call_moncmp(self, ca_moncmp_is_finite); }
  
/*----------------------- match --------------------------*/

static void
ca_bincmp_match_fixlen (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  char *q1 = ptr1, *q2 = ptr2;
  char *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t s1 = b1*i1, s2 = b2*i2, s3 = b3*i3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*s1;
        p2 = q2 + k*s2;
        p3 = q3 + k*s3;
        {
          
         (*p3) = RTEST(rb_funcall(rb_str_new(p1,b1), rb_intern("=~"), 1, (*p2)));
        
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*s1;
      p2=q2+k*s2;
      p3=q3+k*s3;
      {
        
         (*p3) = RTEST(rb_funcall(rb_str_new(p1,b1), rb_intern("=~"), 1, (*p2)));
       
      }
    }    
  }
}

static void
ca_bincmp_match_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = RTEST(rb_funcall((*p1), rb_intern("=~"), 1, (*p2))); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = RTEST(rb_funcall((*p1), rb_intern("=~"), 1, (*p2)));
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_match[CA_NTYPE] = {
  ca_bincmp_match_fixlen,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_match_VALUE,
};


static VALUE rb_ca_match (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("match"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_match);
}
  
/*----------------------- is_kind_of --------------------------*/

static void
ca_bincmp_is_kind_of_VALUE (ca_size_t n, boolean8_t *m, 
                           char *ptr1, ca_size_t b1, ca_size_t i1, 
                           char *ptr2, ca_size_t b2, ca_size_t i2, 
                           char *ptr3, ca_size_t b3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2;
  VALUE *p1 = q1, *p2 = q2;
  boolean8_t *q3 = (boolean8_t *) ptr3;
  boolean8_t *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm = m;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = RTEST(rb_obj_is_kind_of((*p1), (*p2))); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1=q1+k*i1;
      p2=q2+k*i2;
      p3=q3+k*i3;
      {
        (*p3) = RTEST(rb_obj_is_kind_of((*p1), (*p2)));
      }
    }
  }
}

ca_bincmp_func_t
ca_bincmp_is_kind_of[CA_NTYPE] = {
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_not_implement,
  ca_bincmp_is_kind_of_VALUE,
};


static VALUE rb_ca_is_kind_of (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("is_kind_of"));
  }
  return rb_ca_call_bincmp(self, other, ca_bincmp_is_kind_of);
}
  

#define op_powi(type) \
static type \
op_powi_## type (type x, int32_t p) \
{ \
  type r=1; \
\
  switch(p) { \
  case 2: return x*x; \
  case 3: return x*x*x; \
  case 0: return 1; \
  case 1: return x; \
  } \
  if (p<0) { \
    type den = op_powi_## type(x, -p); \
    if (den==0) ca_zerodiv(); \
    return 1/den; \
  }\
  while (p) { \
    if ( (p%2) == 1 ) r *= x; \
    x *= x; \
    p /= 2; \
  } \
  return r; \
}

#define op_powi_fc(type) \
static type \
op_powi_## type (type x, int32_t p) \
{ \
  type r=1; \
\
  switch(p) { \
  case 2: return x*x; \
  case 3: return x*x*x; \
  case 0: return 1; \
  case 1: return x; \
  } \
  if (p<0) { \
    type den = op_powi_## type(x, -p); \
    return 1/den; \
  }\
  while (p) { \
    if ( (p%2) == 1 ) r *= x; \
    x *= x; \
    p /= 2; \
  } \
  return r; \
}

op_powi(int8_t);
op_powi(uint8_t);
op_powi(int16_t);
op_powi(uint16_t);
op_powi(int32_t);
op_powi(uint32_t);
op_powi(int64_t);
op_powi(uint64_t);
op_powi_fc(float32_t);
op_powi_fc(float64_t);
op_powi_fc(float128_t);
op_powi_fc(cmplx64_t);
op_powi_fc(cmplx128_t);

static void
ca_ipower_float32_t (ca_size_t n, boolean8_t *m,
                     char *ptr1, int32_t ipow, char *ptr2)
{
  float32_t *p1 = (float32_t *) ptr1, *p2 = (float32_t *) ptr2;
  if ( m ) {
    while (n--) {
      if ( ! *m++ ) { *p2 = op_powi_float32_t(*p1, ipow); }
      p1++; p2++;
    }
  }
  else
    while (n--) { (*p2) = op_powi_float32_t(*p1, ipow); p1++; p2++; }
}

static void
ca_ipower_float64_t (ca_size_t n, boolean8_t *m,
                     char *ptr1, int32_t ipow, char *ptr2)
{
  float64_t *p1 = (float64_t *) ptr1, *p2 = (float64_t *) ptr2;
  if ( m ) {
    while (n--) {
      if ( ! *m++ ) { *p2 = op_powi_float64_t(*p1, ipow); }
      p1++; p2++;
    }
  }
  else
    while (n--) { (*p2) = op_powi_float64_t(*p1, ipow); p1++; p2++; }
}

static void
ca_ipower_float128_t (ca_size_t n, boolean8_t *m,
                     char *ptr1, int32_t ipow, char *ptr2)
{
  float128_t *p1 = (float128_t *) ptr1, *p2 = (float128_t *) ptr2;
  if ( m ) {
    while (n--) {
      if ( ! *m++ ) { *p2 = op_powi_float128_t(*p1, ipow); }
      p1++; p2++;
    }
  }
  else
    while (n--) { (*p2) = op_powi_float128_t(*p1, ipow); p1++; p2++; }
}

static void
ca_ipower_cmplx64_t (ca_size_t n, boolean8_t *m,
                     char *ptr1, int32_t ipow, char *ptr2)
{
  cmplx64_t *p1 = (cmplx64_t *) ptr1, *p2 = (cmplx64_t *) ptr2;
  if ( m ) {
    while (n--) {
      if ( ! *m++ ) { *p2 = op_powi_cmplx64_t(*p1, ipow); }
      p1++; p2++;
    }
  }
  else
    while (n--) { (*p2) = op_powi_cmplx64_t(*p1, ipow); p1++; p2++; }
}

static void
ca_ipower_cmplx128_t (ca_size_t n, boolean8_t *m,
                     char *ptr1, int32_t ipow, char *ptr2)
{
  cmplx128_t *p1 = (cmplx128_t *) ptr1, *p2 = (cmplx128_t *) ptr2;
  if ( m ) {
    while (n--) {
      if ( ! *m++ ) { *p2 = op_powi_cmplx128_t(*p1, ipow); }
      p1++; p2++;
    }
  }
  else
    while (n--) { (*p2) = op_powi_cmplx128_t(*p1, ipow); p1++; p2++; }
}


static VALUE
rb_ca_ipower (VALUE self, VALUE other)
{
  volatile VALUE obj;
  CArray *ca, *co;
  int32_t ipow;
  boolean8_t *m;

  ipow = NUM2INT(other);

  Data_Get_Struct(self, CArray, ca);

  co = ca_template(ca);
  obj = ca_wrap_struct(co);

  ca_attach(ca);

  ca_copy_mask_overlay(co, co->elements, 1, ca);
  m = ( co->mask ) ? (boolean8_t *)co->mask->ptr : NULL;

  switch ( ca->data_type ) {
  case CA_FLOAT32: {
    ca_ipower_float32_t(ca->elements, m, ca->ptr, ipow, co->ptr);
    break;
  }
  case CA_FLOAT64: {
    ca_ipower_float64_t(ca->elements, m, ca->ptr, ipow, co->ptr);
    break;
  }
  case CA_FLOAT128: {
    ca_ipower_float128_t(ca->elements, m, ca->ptr, ipow, co->ptr);
    break;
  }
  case CA_CMPLX64: {
    ca_ipower_cmplx64_t(ca->elements, m, ca->ptr, ipow, co->ptr);
    break;
  }
  case CA_CMPLX128: {
    ca_ipower_cmplx128_t(ca->elements, m, ca->ptr, ipow, co->ptr);
    break;
  }
  default:
    rb_raise(rb_eCADataTypeError, "invalid data type for ipower");
    break;
  }

  ca_detach(ca);

  /* unresolved unbound repeat array generates unbound repeat array again */
  if ( ca->obj_type == CA_OBJ_UNBOUND_REPEAT ) {
    CAUnboundRepeat *cx = (CAUnboundRepeat *) ca;
    obj = rb_ca_ubrep_new(rb_ca_ubrep_shave(self, obj), cx->rep_ndim, cx->rep_dim);
  }

  return obj;
}

static VALUE
rb_ca_ipower_bang (VALUE self, VALUE other)
{
  CArray *ca;
  int32_t ipow;
  boolean8_t *m;

  ipow = NUM2INT(other);

  Data_Get_Struct(self, CArray, ca);

  ca_attach(ca);

  m = ( ca->mask ) ? (boolean8_t *)ca->mask->ptr : NULL;

  switch ( ca->data_type ) {
  case CA_FLOAT32: {
    ca_ipower_float32_t(ca->elements, m, ca->ptr, ipow, ca->ptr);
    break;
  }
  case CA_FLOAT64: {
    ca_ipower_float64_t(ca->elements, m, ca->ptr, ipow, ca->ptr);
    break;
  }
  case CA_FLOAT128: {
    ca_ipower_float128_t(ca->elements, m, ca->ptr, ipow, ca->ptr);
    break;
  }
  case CA_CMPLX64: {
    ca_ipower_cmplx64_t(ca->elements, m, ca->ptr, ipow, ca->ptr);
    break;
  }
  case CA_CMPLX128: {
    ca_ipower_cmplx128_t(ca->elements, m, ca->ptr, ipow, ca->ptr);
    break;
  }
  default:
    rb_raise(rb_eRuntimeError, "invalid data type for ipower");
    break;
  }

  ca_sync(ca);
  ca_detach(ca);

  return self;
}


/*----------------------- power --------------------------*/

static void
ca_binop_power_int8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int8_t *q1 = (int8_t *) ptr1, *q2 = (int8_t *) ptr2, *q3 = (int8_t *) ptr3;
  int8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_int8_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_int8_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_uint8_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint8_t *q1 = (uint8_t *) ptr1, *q2 = (uint8_t *) ptr2, *q3 = (uint8_t *) ptr3;
  uint8_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_uint8_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_uint8_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_int16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int16_t *q1 = (int16_t *) ptr1, *q2 = (int16_t *) ptr2, *q3 = (int16_t *) ptr3;
  int16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_int16_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_int16_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_uint16_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint16_t *q1 = (uint16_t *) ptr1, *q2 = (uint16_t *) ptr2, *q3 = (uint16_t *) ptr3;
  uint16_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_uint16_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_uint16_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_int32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int32_t *q1 = (int32_t *) ptr1, *q2 = (int32_t *) ptr2, *q3 = (int32_t *) ptr3;
  int32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_int32_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_int32_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_uint32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint32_t *q1 = (uint32_t *) ptr1, *q2 = (uint32_t *) ptr2, *q3 = (uint32_t *) ptr3;
  uint32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_uint32_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_uint32_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_int64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  int64_t *q1 = (int64_t *) ptr1, *q2 = (int64_t *) ptr2, *q3 = (int64_t *) ptr3;
  int64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_int64_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_int64_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_uint64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  uint64_t *q1 = (uint64_t *) ptr1, *q2 = (uint64_t *) ptr2, *q3 = (uint64_t *) ptr3;
  uint64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = op_powi_uint64_t((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = op_powi_uint64_t((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_float32_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float32_t *q1 = (float32_t *) ptr1, *q2 = (float32_t *) ptr2, *q3 = (float32_t *) ptr3;
  float32_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = pow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = pow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_float64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float64_t *q1 = (float64_t *) ptr1, *q2 = (float64_t *) ptr2, *q3 = (float64_t *) ptr3;
  float64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = pow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = pow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_float128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  float128_t *q1 = (float128_t *) ptr1, *q2 = (float128_t *) ptr2, *q3 = (float128_t *) ptr3;
  float128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = pow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = pow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_cmplx64_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx64_t *q1 = (cmplx64_t *) ptr1, *q2 = (cmplx64_t *) ptr2, *q3 = (cmplx64_t *) ptr3;
  cmplx64_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = cpow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = cpow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_cmplx128_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx128_t *q1 = (cmplx128_t *) ptr1, *q2 = (cmplx128_t *) ptr2, *q3 = (cmplx128_t *) ptr3;
  cmplx128_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = cpow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = cpow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_cmplx256_t (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  cmplx256_t *q1 = (cmplx256_t *) ptr1, *q2 = (cmplx256_t *) ptr2, *q3 = (cmplx256_t *) ptr3;
  cmplx256_t *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = cpow((*p1), (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 1
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = cpow((*p1), (*p2));
      }
    }
  }
}

static void
ca_binop_power_VALUE (ca_size_t n, boolean8_t *m, char *ptr1, ca_size_t i1, char *ptr2, ca_size_t i2, char *ptr3, ca_size_t i3)
{
  VALUE *q1 = (VALUE *) ptr1, *q2 = (VALUE *) ptr2, *q3 = (VALUE *) ptr3;
  VALUE *p1 = q1, *p2 = q2, *p3 = q3;
  ca_size_t k;
  if ( m ) {
    boolean8_t *pm;
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(pm,p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      pm = m + k;
      if ( ! *pm ) {
        p1 = q1 + k*i1;
        p2 = q2 + k*i2;
        p3 = q3 + k*i3;
        {
          (*p3) = rb_funcall((*p1), id_star_star, 1, (*p2)); 
        }
      }
    }
  }
  else {
    #if defined(_OPENMP) && 0
    #pragma omp parallel for private(p1,p2,p3)
    #endif
    for (k=0; k<n; k++) {
      p1 = q1 + k*i1;
      p2 = q2 + k*i2;
      p3 = q3 + k*i3;
      {
        (*p3) = rb_funcall((*p1), id_star_star, 1, (*p2));
      }
    }
  }
}

ca_binop_func_t
ca_binop_power[CA_NTYPE] = {
  ca_binop_not_implement,
  ca_binop_not_implement,
  ca_binop_power_int8_t,
  ca_binop_power_uint8_t,
  ca_binop_power_int16_t,
  ca_binop_power_uint16_t,
  ca_binop_power_int32_t,
  ca_binop_power_uint32_t,
  ca_binop_power_int64_t,
  ca_binop_power_uint64_t,
  ca_binop_power_float32_t,
  ca_binop_power_float64_t,
  ca_binop_power_float128_t,
  ca_binop_power_cmplx64_t,
  ca_binop_power_cmplx128_t,
  ca_binop_power_cmplx256_t,
  ca_binop_power_VALUE,
};


static VALUE rb_ca_power (VALUE self, VALUE other)
{ 
  if ( ! rb_ca_test_castable(other) ) {
    return rb_ca_binop_pass_to_other(self, other, rb_intern("power"));
  }
  return rb_ca_call_binop(self, other, ca_binop_power); 
}
static VALUE rb_ca_power_bang (VALUE self, VALUE other)
{ return rb_ca_call_binop_bang(self, other, ca_binop_power); }
  

static VALUE rb_ca_pow (VALUE self, VALUE other)
{
  volatile VALUE obj;
  CArray *ca;

  Data_Get_Struct(self, CArray, ca);
  if ( ( ca_is_float_type(ca) || ca_is_complex_type(ca) ) &&
       rb_obj_is_kind_of(other, rb_cInteger) ) {
    return rb_ca_ipower(self, other);
  }
  else {
    obj = rb_ca_power(self, other);

    /* unresolved unbound repeat array generates unbound repeat array again */
    if ( ca->obj_type == CA_OBJ_UNBOUND_REPEAT ) {
      CAUnboundRepeat *cx = (CAUnboundRepeat *) ca;      
      obj = rb_ca_ubrep_new(rb_ca_ubrep_shave(self, obj), cx->rep_ndim, cx->rep_dim);
    }

    return obj;
  }
}

static VALUE rb_ca_pow_bang (VALUE self, VALUE other)
{
  CArray *ca;
  Data_Get_Struct(self, CArray, ca);
  if ( ( ca_is_float_type(ca) || ca_is_complex_type(ca) ) &&
       rb_obj_is_kind_of(other, rb_cInteger) ) {
    return rb_ca_ipower_bang(self, other);
  }
  else {
    return rb_ca_power_bang(self, other);
  }
}



void
Init_carray_math ()
{
  
  rb_define_method(rb_cCArray, "zero", rb_ca_zero, 0);

  rb_define_method(rb_cCArray, "zero!", rb_ca_zero_bang, 0);

  rb_define_method(rb_cCArray, "one", rb_ca_one, 0);

  rb_define_method(rb_cCArray, "one!", rb_ca_one_bang, 0);

  rb_define_method(rb_cCArray, "frac", rb_ca_frac, 0);

  rb_define_method(rb_cCArray, "frac!", rb_ca_frac_bang, 0);

  rb_define_method(rb_cCArray, "neg", rb_ca_neg, 0);

  rb_define_method(rb_cCArray, "neg!", rb_ca_neg_bang, 0);

  rb_define_method(rb_cCArray, "bit_neg", rb_ca_bit_neg, 0);

  rb_define_method(rb_cCArray, "bit_neg!", rb_ca_bit_neg_bang, 0);

  rb_define_alias(rb_cCArray, "-@", "neg");

  rb_define_alias(rb_cCArray, "~", "bit_neg");

  rb_define_method(rb_cCArray, "abs_i", rb_ca_abs_i, 0);

  rb_define_method(rb_cCArray, "abs_i!", rb_ca_abs_i_bang, 0);

  rb_define_method(rb_cCArray, "abs", rb_ca_abs, 0);
  rb_define_method(rb_cCArray, "abs!", rb_ca_abs_bang, 0);
  rb_define_module_function(rb_mCAMath, "abs", rb_cmath_abs, 1);

  rb_define_method(rb_cCArray, "conj", rb_ca_conj, 0);

  rb_define_method(rb_cCArray, "conj!", rb_ca_conj_bang, 0);

  rb_define_method(rb_cCArray, "rad", rb_ca_rad, 0);

  rb_define_method(rb_cCArray, "rad!", rb_ca_rad_bang, 0);

  rb_define_module_function(rb_mCAMath, "rad", rb_cmath_rad, 1);

  rb_define_method(rb_cCArray, "deg", rb_ca_deg, 0);

  rb_define_method(rb_cCArray, "deg!", rb_ca_deg_bang, 0);

  rb_define_module_function(rb_mCAMath, "deg", rb_cmath_deg, 1);

  rb_define_method(rb_cCArray, "ceil", rb_ca_ceil, 0);

  rb_define_method(rb_cCArray, "ceil!", rb_ca_ceil_bang, 0);

  rb_define_module_function(rb_mCAMath, "ceil", rb_cmath_ceil, 1);

  rb_define_method(rb_cCArray, "floor", rb_ca_floor, 0);

  rb_define_method(rb_cCArray, "floor!", rb_ca_floor_bang, 0);

  rb_define_module_function(rb_mCAMath, "floor", rb_cmath_floor, 1);

  rb_define_method(rb_cCArray, "round", rb_ca_round, 0);

  rb_define_method(rb_cCArray, "round!", rb_ca_round_bang, 0);

  rb_define_module_function(rb_mCAMath, "round", rb_cmath_round, 1);

  rb_define_method(rb_cCArray, "rcp", rb_ca_rcp, 0);

  rb_define_method(rb_cCArray, "rcp!", rb_ca_rcp_bang, 0);

  rb_define_module_function(rb_mCAMath, "rcp", rb_cmath_rcp, 1);

  rb_define_method(rb_cCArray, "sqrt", rb_ca_sqrt, 0);

  rb_define_method(rb_cCArray, "sqrt!", rb_ca_sqrt_bang, 0);

  rb_define_module_function(rb_mCAMath, "sqrt", rb_cmath_sqrt, 1);

  rb_define_method(rb_cCArray, "exp", rb_ca_exp, 0);

  rb_define_method(rb_cCArray, "exp!", rb_ca_exp_bang, 0);

  rb_define_module_function(rb_mCAMath, "exp", rb_cmath_exp, 1);

  rb_define_method(rb_cCArray, "exp2", rb_ca_exp2, 0);

  rb_define_method(rb_cCArray, "exp2!", rb_ca_exp2_bang, 0);

  rb_define_module_function(rb_mCAMath, "exp2", rb_cmath_exp2, 1);

  rb_define_method(rb_cCArray, "exp10", rb_ca_exp10, 0);

  rb_define_method(rb_cCArray, "exp10!", rb_ca_exp10_bang, 0);

  rb_define_module_function(rb_mCAMath, "exp10", rb_cmath_exp10, 1);

  rb_define_method(rb_cCArray, "log", rb_ca_log, 0);

  rb_define_method(rb_cCArray, "log!", rb_ca_log_bang, 0);

  rb_define_module_function(rb_mCAMath, "log", rb_cmath_log, 1);

  rb_define_method(rb_cCArray, "log10", rb_ca_log10, 0);

  rb_define_method(rb_cCArray, "log10!", rb_ca_log10_bang, 0);

  rb_define_module_function(rb_mCAMath, "log10", rb_cmath_log10, 1);

  rb_define_method(rb_cCArray, "log2", rb_ca_log2, 0);

  rb_define_method(rb_cCArray, "log2!", rb_ca_log2_bang, 0);

  rb_define_module_function(rb_mCAMath, "log2", rb_cmath_log2, 1);

  rb_define_method(rb_cCArray, "logb", rb_ca_logb, 0);

  rb_define_method(rb_cCArray, "logb!", rb_ca_logb_bang, 0);

  rb_define_module_function(rb_mCAMath, "logb", rb_cmath_logb, 1);

  rb_define_method(rb_cCArray, "sin", rb_ca_sin, 0);

  rb_define_method(rb_cCArray, "sin!", rb_ca_sin_bang, 0);

  rb_define_module_function(rb_mCAMath, "sin", rb_cmath_sin, 1);

  rb_define_method(rb_cCArray, "cos", rb_ca_cos, 0);

  rb_define_method(rb_cCArray, "cos!", rb_ca_cos_bang, 0);

  rb_define_module_function(rb_mCAMath, "cos", rb_cmath_cos, 1);

  rb_define_method(rb_cCArray, "tan", rb_ca_tan, 0);

  rb_define_method(rb_cCArray, "tan!", rb_ca_tan_bang, 0);

  rb_define_module_function(rb_mCAMath, "tan", rb_cmath_tan, 1);

  rb_define_method(rb_cCArray, "asin", rb_ca_asin, 0);

  rb_define_method(rb_cCArray, "asin!", rb_ca_asin_bang, 0);

  rb_define_module_function(rb_mCAMath, "asin", rb_cmath_asin, 1);

  rb_define_method(rb_cCArray, "acos", rb_ca_acos, 0);

  rb_define_method(rb_cCArray, "acos!", rb_ca_acos_bang, 0);

  rb_define_module_function(rb_mCAMath, "acos", rb_cmath_acos, 1);

  rb_define_method(rb_cCArray, "atan", rb_ca_atan, 0);

  rb_define_method(rb_cCArray, "atan!", rb_ca_atan_bang, 0);

  rb_define_module_function(rb_mCAMath, "atan", rb_cmath_atan, 1);

  rb_define_method(rb_cCArray, "sinh", rb_ca_sinh, 0);

  rb_define_method(rb_cCArray, "sinh!", rb_ca_sinh_bang, 0);

  rb_define_module_function(rb_mCAMath, "sinh", rb_cmath_sinh, 1);

  rb_define_method(rb_cCArray, "cosh", rb_ca_cosh, 0);

  rb_define_method(rb_cCArray, "cosh!", rb_ca_cosh_bang, 0);

  rb_define_module_function(rb_mCAMath, "cosh", rb_cmath_cosh, 1);

  rb_define_method(rb_cCArray, "tanh", rb_ca_tanh, 0);

  rb_define_method(rb_cCArray, "tanh!", rb_ca_tanh_bang, 0);

  rb_define_module_function(rb_mCAMath, "tanh", rb_cmath_tanh, 1);

  rb_define_method(rb_cCArray, "asinh", rb_ca_asinh, 0);

  rb_define_method(rb_cCArray, "asinh!", rb_ca_asinh_bang, 0);

  rb_define_module_function(rb_mCAMath, "asinh", rb_cmath_asinh, 1);

  rb_define_method(rb_cCArray, "acosh", rb_ca_acosh, 0);

  rb_define_method(rb_cCArray, "acosh!", rb_ca_acosh_bang, 0);

  rb_define_module_function(rb_mCAMath, "acosh", rb_cmath_acosh, 1);

  rb_define_method(rb_cCArray, "atanh", rb_ca_atanh, 0);

  rb_define_method(rb_cCArray, "atanh!", rb_ca_atanh_bang, 0);

  rb_define_module_function(rb_mCAMath, "atanh", rb_cmath_atanh, 1);

  rb_define_method(rb_cCArray, "pmax", rb_ca_pmax, 1);

  rb_define_method(rb_cCArray, "pmax!", rb_ca_pmax_bang, 1);

  rb_define_method(rb_cCArray, "pmin", rb_ca_pmin, 1);

  rb_define_method(rb_cCArray, "pmin!", rb_ca_pmin_bang, 1);

  rb_define_method(rb_cCArray, "+", rb_ca_add, 1);

  rb_define_method(rb_cCArray, "add!", rb_ca_add_bang, 1);

  rb_define_method(rb_cCArray, "-", rb_ca_sub, 1);

  rb_define_method(rb_cCArray, "sub!", rb_ca_sub_bang, 1);

  rb_define_method(rb_cCArray, "*", rb_ca_mul, 1);

  rb_define_method(rb_cCArray, "mul!", rb_ca_mul_bang, 1);

  rb_define_method(rb_cCArray, "/", rb_ca_div, 1);

  rb_define_method(rb_cCArray, "div!", rb_ca_div_bang, 1);

  rb_define_method(rb_cCArray, "quo_i", rb_ca_quo_i, 1);

  rb_define_method(rb_cCArray, "quo_i!", rb_ca_quo_i_bang, 1);

  rb_define_method(rb_cCArray, "rcp_mul", rb_ca_rcp_mul, 1);

  rb_define_method(rb_cCArray, "rcp_mul!", rb_ca_rcp_mul_bang, 1);

  rb_define_method(rb_cCArray, "%", rb_ca_mod, 1);

  rb_define_method(rb_cCArray, "mod!", rb_ca_mod_bang, 1);

  rb_define_method(rb_cCArray, "reminder", rb_ca_reminder, 1);

  rb_define_method(rb_cCArray, "reminder!", rb_ca_reminder_bang, 1);

  rb_define_method(rb_cCArray, "&", rb_ca_bit_and_i, 1);

  rb_define_method(rb_cCArray, "bit_and_i!", rb_ca_bit_and_i_bang, 1);

  rb_define_method(rb_cCArray, "|", rb_ca_bit_or_i, 1);

  rb_define_method(rb_cCArray, "bit_or_i!", rb_ca_bit_or_i_bang, 1);

  rb_define_method(rb_cCArray, "^", rb_ca_bit_xor_i, 1);

  rb_define_method(rb_cCArray, "bit_xor_i!", rb_ca_bit_xor_i_bang, 1);

  rb_define_method(rb_cCArray, "bit_and", rb_ca_bit_and, 1);
  rb_define_method(rb_cCArray, "bit_or", rb_ca_bit_or, 1);
  rb_define_method(rb_cCArray, "bit_xor", rb_ca_bit_xor, 1);

  rb_define_method(rb_cCArray, "<<", rb_ca_bit_lshift, 1);

  rb_define_method(rb_cCArray, "bit_lshift!", rb_ca_bit_lshift_bang, 1);

  rb_define_method(rb_cCArray, ">>", rb_ca_bit_rshift, 1);

  rb_define_method(rb_cCArray, "bit_rshift!", rb_ca_bit_rshift_bang, 1);

  rb_define_method(rb_cCArray, "feq", rb_ca_feq, 1);

  rb_define_method(rb_cCArray, "eq", rb_ca_eq, 1);

  rb_define_method(rb_cCArray, "ne", rb_ca_ne, 1);

  rb_define_method(rb_cCArray, "gt", rb_ca_gt, 1);

  rb_define_method(rb_cCArray, "lt", rb_ca_lt, 1);

  rb_define_method(rb_cCArray, "ge", rb_ca_ge, 1);

  rb_define_method(rb_cCArray, "le", rb_ca_le, 1);

  rb_define_method(rb_cCArray, "not", rb_ca_not, 0);

  rb_define_method(rb_cCArray, "not!", rb_ca_not_bang, 0);

  rb_define_method(rb_cCArray, "and", rb_ca_and_i, 1);

  rb_define_method(rb_cCArray, "and_i!", rb_ca_and_i_bang, 1);

  rb_define_method(rb_cCArray, "or", rb_ca_or_i, 1);

  rb_define_method(rb_cCArray, "or_i!", rb_ca_or_i_bang, 1);

  rb_define_method(rb_cCArray, "xor", rb_ca_xor_i, 1);

  rb_define_method(rb_cCArray, "xor_i!", rb_ca_xor_i_bang, 1);

  rb_define_method(rb_cCArray, "and", rb_ca_and, 1);
  rb_define_method(rb_cCArray, "or", rb_ca_or, 1);
  rb_define_method(rb_cCArray, "xor", rb_ca_xor, 1);

  rb_define_alias(rb_cCArray, "add", "+");

  rb_define_alias(rb_cCArray, "sub", "-");

  rb_define_alias(rb_cCArray, "mul", "*");

  rb_define_alias(rb_cCArray, "div", "/");

  rb_define_alias(rb_cCArray, "mod", "%");

  rb_define_alias(rb_cCArray, "bit_and", "&");

  rb_define_alias(rb_cCArray, "bit_or", "|");

  rb_define_alias(rb_cCArray, "bit_xor", "^");

  rb_define_alias(rb_cCArray, "bit_lshift", "<<");

  rb_define_alias(rb_cCArray, "bit_rshift", ">>");

  rb_define_alias(rb_cCArray, ">", "gt");

  rb_define_alias(rb_cCArray, "<", "lt");

  rb_define_alias(rb_cCArray, ">=", "ge");

  rb_define_alias(rb_cCArray, "<=", "le");

  rb_define_method(rb_cCArray, "is_nan", rb_ca_is_nan, 0);

  rb_define_method(rb_cCArray, "is_inf", rb_ca_is_inf, 0);

  rb_define_method(rb_cCArray, "is_finite", rb_ca_is_finite, 0);

  rb_define_method(rb_cCArray, "match", rb_ca_match, 1);

  rb_define_method(rb_cCArray, "is_kind_of", rb_ca_is_kind_of, 1);

  rb_define_alias(rb_cCArray, "=~", "match");

  rb_define_method(rb_cCArray, "power", rb_ca_power, 1);

  rb_define_method(rb_cCArray, "power!", rb_ca_power_bang, 1);

  rb_define_method(rb_cCArray, "pow", rb_ca_pow, 1);
  rb_define_method(rb_cCArray, "pow!", rb_ca_pow_bang, 1);

  rb_define_alias(rb_cCArray, "**", "pow");

  id_equal = rb_intern("==");
  id_lt    = rb_intern("<");
  id_le    = rb_intern("<=");
  id_gt    = rb_intern(">");
  id_ge    = rb_intern(">=");

  id_uminus = rb_intern("-@");
  id_utilda = rb_intern("~@");

  id_plus  = rb_intern("+");
  id_minus = rb_intern("-");
  id_star  = rb_intern("*");
  id_slash = rb_intern("/");
  id_percent = rb_intern("%");
  id_star_star = rb_intern("**");

  id_and  = rb_intern("&");
  id_or   = rb_intern("|");
  id_xor  = rb_intern("^");
  id_lshift = rb_intern("<<");
  id_rshift = rb_intern(">>");


}

