/* ---------------------------------------------------------------------------

  carray_stat_proc.c

  This file is part of Ruby/CArray extension library.


  This file is automatically generated from carray_stat_proc.rb
  Copyright (C) 2005-2020 Hiroki Motoyoshi

---------------------------------------------------------------------------- */

#include <math.h>
#include "ruby.h"
#include "carray.h"

typedef struct {
  ca_size_t    offset;
  ca_size_t    count;
  ca_size_t    step;
  ca_size_t   *addr;
} CAStatIterator;

typedef void (*ca_stat_proc_t)();

#define iterator_rewind(it)                     {     if ( (it)->step ) {       (it)->count = 0;       (it)->offset = (it)->addr[((it)->count)++];     }     else {       (it)->offset = 0;     }   }

#define iterator_succ(it)                       {     if ( (it)->step ) {       (it)->offset = (it)->addr[((it)->count)++];     }     else {       ((it)->offset)++;     }   }

static ID id_lt, id_gt, id_plus, id_minus, id_star, id_quo;

#define GEQ(x,y)        ( (x) == (y) )
#define OEQ(x,y)        rb_equal((x),(y))

#define lt_c(x, y)      ( (x) < (y) )
#define lt_VALUE(x, y)  rb_funcall((x), id_lt, 1, (y))

#define gt_c(x, y)      ( (x) > (y) )
#define gt_VALUE(x, y)  rb_funcall((x), id_gt, 1, (y))

#define add_c(x, y)     ( (x) + (y) )
#define add_VALUE(x, y) rb_funcall((x), id_plus, 1, (y))

#define sub_c(x, y)     ( (x) - (y) )
#define sub_VALUE(x, y) rb_funcall((x), id_minus, 1, (y))

#define mul_c(x, y)     ( (x) * (y) )
#define mul_VALUE(x, y) rb_funcall((x), id_star, 1, (y))

#define div_c(x, y)     ( (x) / (y) )
#define div_VALUE(x, y) rb_funcall((x), id_quo, 1, (y))

#define sqrt_c(x)       sqrt(x)
#define sqrt_VALUE(x)   rb_funcall((x), rb_intern("sqrt"), 0)


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_boolean8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_boolean8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_boolean8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_boolean8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, boolean8_t *retval)
{
  boolean8_t min = 0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_boolean8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  boolean8_t min = 0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_boolean8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, boolean8_t *retval)
{
  boolean8_t max = 0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_boolean8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  boolean8_t max = 0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_boolean8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_boolean8_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_int8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_int8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_int8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_int8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int8_t *retval)
{
  int8_t min = 0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_int8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int8_t min = 0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_int8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int8_t *retval)
{
  int8_t max = 0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_int8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int8_t max = 0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_int8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_int8_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_uint8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_uint8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_uint8_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_uint8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint8_t *retval)
{
  uint8_t min = 0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_uint8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint8_t min = 0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_uint8_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint8_t *retval)
{
  uint8_t max = 0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_uint8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint8_t max = 0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_uint8_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_uint8_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_int16_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_int16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_int16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_int16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int16_t *retval)
{
  int16_t min = 0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_int16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int16_t min = 0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_int16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int16_t *retval)
{
  int16_t max = 0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_int16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int16_t max = 0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_int16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_int16_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_uint16_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_uint16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_uint16_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_uint16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint16_t *retval)
{
  uint16_t min = 0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_uint16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint16_t min = 0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_uint16_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint16_t *retval)
{
  uint16_t max = 0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_uint16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint16_t max = 0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_uint16_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_uint16_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_int32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_int32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_int32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_int32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int32_t *retval)
{
  int32_t min = 0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_int32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int32_t min = 0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_int32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int32_t *retval)
{
  int32_t max = 0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_int32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int32_t max = 0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_int32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_int32_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_uint32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_uint32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_uint32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_uint32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint32_t *retval)
{
  uint32_t min = 0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_uint32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint32_t min = 0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_uint32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint32_t *retval)
{
  uint32_t max = 0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_uint32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint32_t max = 0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_uint32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_uint32_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_int64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_int64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_int64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_int64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int64_t *retval)
{
  int64_t min = 0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LL2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_int64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int64_t min = 0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_int64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, int64_t *retval)
{
  int64_t max = 0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LL2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_int64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  int64_t max = 0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_int64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_int64_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_uint64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_uint64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_uint64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_uint64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint64_t *retval)
{
  uint64_t min = 0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULL2NUM(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_uint64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint64_t min = 0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_uint64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, uint64_t *retval)
{
  uint64_t max = 0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULL2NUM(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_uint64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  uint64_t max = 0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_uint64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_uint64_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_float32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_float32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (float64_t)(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (float64_t)(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_float32_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_float32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float32_t *retval)
{
  float32_t min = 0.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_float32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float32_t min = 0.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_float32_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float32_t *retval)
{
  float32_t max = 0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_float32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float32_t max = 0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_float32_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_float32_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (float64_t)(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (float64_t)(sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_float64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_float64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_float64_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_float64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  float64_t min = 0.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_float64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float64_t min = 0.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_float64_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  float64_t max = 0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_float64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float64_t max = 0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_float64_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_float64_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double prod = 1.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_c(prod, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_c(prod, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_float128_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_float128_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = (sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, ave;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (sqrt_c(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = (var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_float128_t (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, long double *retval)
{
  volatile long double sum = 0.0, sum2 = 0.0, ave, var, diff;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_c(sum, (long double)(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_c((long double)(*(p + *a)), ave);
        sum2 = add_c(sum2, mul_c(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_c((long double)(*(p + *a)), ave);
      sum2 = add_c(sum2, mul_c(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_c(sum2, (long double)(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sqrt_c(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_c(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_float128_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float128_t *retval)
{
  float128_t min = 0.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_float128_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float128_t min = 0.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_c(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_c(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_float128_t (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float128_t *retval)
{
  float128_t max = 0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_float128_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  float128_t max = 0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_c(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_c(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_float128_t (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_float128_t (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile long double sum = 0.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (long double)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = (sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (long double)(*(p + *a)));
      *retval = (sum);
      retval++;
      iterator_succ(it);
    }
  }
}


/* ============================= */
/* ca_proc_prod                  */
/* ============================= */

static void
ca_proc_prod_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE prod = rb_float_new(1.0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        prod = mul_VALUE(prod, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      prod = mul_VALUE(prod, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (prod);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = NUM2DBL(prod);
  }
}

/* ============================= */
/* ca_proc_count                 */
/* ============================= */

static void
ca_proc_count_VALUE (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : SIZE2NUM(elements - count);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = elements - count;
  }
}

static void
ca_proc_sum_VALUE (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *)it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = NUM2DBL(sum);
  }
}


/* ============================= */
/* ca_proc_mean                  */
/* ============================= */

static void
ca_proc_mean_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0), ave;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_VALUE(sum, (VALUE)INT2NUM(elements-count));
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (ave);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = NUM2DBL(ave);
  }
}

/* ============================= */
/* ca_proc_variancep             */
/* ============================= */

static void
ca_proc_variancep_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0), sum2 = rb_float_new(0.0), ave, var, diff;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_VALUE(sum, (VALUE)INT2NUM(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_VALUE((VALUE)(*(p + *a)), ave);
        sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_VALUE((VALUE)(*(p + *a)), ave);
      sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_VALUE(sum2, (VALUE)INT2NUM(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = NUM2DBL(var);
  }
}

/* ============================= */
/* ca_proc_stddevp             */
/* ============================= */

static void
ca_proc_stddevp_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0), sum2 = rb_float_new(0.0), ave, var, diff;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_VALUE(sum, (VALUE)INT2NUM(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_VALUE((VALUE)(*(p + *a)), ave);
        sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_VALUE((VALUE)(*(p + *a)), ave);
      sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_VALUE(sum2, (VALUE)INT2NUM(elements-count));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (sqrt_VALUE(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = NUM2DBL(sqrt_VALUE(var));
  }
}

/* ============================= */
/* ca_proc_variance              */
/* ============================= */

static void
ca_proc_variance_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0), sum2 = rb_float_new(0.0), ave, var, diff;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_VALUE(sum, (VALUE)INT2NUM(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_VALUE((VALUE)(*(p + *a)), ave);
        sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_VALUE((VALUE)(*(p + *a)), ave);
      sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_VALUE(sum2, (VALUE)INT2NUM(elements-count-1));

  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (var);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval = NUM2DBL(var);
  }
}

/* ============================= */
/* ca_proc_stddev                */
/* ============================= */

static void
ca_proc_stddev_VALUE (ca_size_t elements, ca_size_t min_count,
                     boolean8_t *m, void *ptr, CAStatIterator *it,
                     int return_object, VALUE *retobj,
                     boolean8_t *retmask, VALUE *retval)
{
  volatile VALUE sum = rb_float_new(0.0), sum2 = rb_float_new(0.0), ave, var, diff;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  ave = div_VALUE(sum, (VALUE)INT2NUM(elements-count));

  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        ;
      }
      else {
        diff = sub_VALUE((VALUE)(*(p + *a)), ave);
        sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      diff = sub_VALUE((VALUE)(*(p + *a)), ave);
      sum2 = add_VALUE(sum2, mul_VALUE(diff,diff));
      iterator_succ(it);
    }
  }

  var = div_VALUE(sum2, (VALUE)INT2NUM(elements-count-1));
  
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (sqrt_VALUE(var));
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sqrt_VALUE(var);
  }
}

/* ============================= */
/* ca_proc_min                   */
/* ============================= */

static void
ca_proc_min_VALUE (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, VALUE *retval)
{
  VALUE min = INT2NUM(0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_VALUE(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_VALUE(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (min);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = min;
  }
}

/* ============================= */
/* ca_proc_min_addr              */
/* ============================= */

static void
ca_proc_min_addr_VALUE (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  VALUE min = INT2NUM(0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        min = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( lt_VALUE(*(p + *a), min) ) {
            min = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    min = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( lt_VALUE(*(p + *a), min) ) {
        min = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}

/* ============================= */
/* ca_proc_max                   */
/* ============================= */

static void
ca_proc_max_VALUE (ca_size_t elements, ca_size_t min_count,
                    boolean8_t *m, void *ptr, CAStatIterator *it,
                    int return_object, VALUE *retobj,
                    boolean8_t *retmask, VALUE *retval)
{
  VALUE max = 0;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_VALUE(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_VALUE(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (max);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = max;
  }
}

/* ============================= */
/* ca_proc_max_addr              */
/* ============================= */

static void
ca_proc_max_addr_VALUE (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         int return_object, VALUE *retobj,
                         boolean8_t *retmask, ca_size_t *retval)
{
  VALUE max = 0;
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t addr;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    addr = -1;
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        max = *(p + *a);
        addr = i;
        break;
      }
      iterator_succ(it);
    }
    iterator_succ(it);
    if ( addr >= 0 ) {
      for (i=addr+1; i<elements; i++) {
        if ( *(m + *a) ) {
          count++;
        }
        else {
          if ( gt_VALUE(*(p + *a), max) ) {
            max = *(p + *a);
            addr = i;
          }
        }
        iterator_succ(it);
      }
    }
  }
  else {
    max = *(p + *a);
    addr = 0;
    for (i=addr; i<elements; i++) {
      if ( gt_VALUE(*(p + *a), max) ) {
        max = *(p + *a);
        addr = i;
      }
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF
                                    : ( addr < 0 ) ? Qnil : SIZE2NUM(addr);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = addr;
  }
}


/* ============================= */
/* ca_proc_cumcount              */
/* ============================= */

static void
ca_proc_cumcount_VALUE (ca_size_t elements, ca_size_t min_count,
                         boolean8_t *m, void *ptr, CAStatIterator *it,
                         boolean8_t *retmask, ca_size_t *retval)
{
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval  = i + 1 - count;
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      *retval = i+1;
      retval++;
    }
  }
}


/* ============================= */
/* ca_proc_cumsum                */
/* ============================= */

static void
ca_proc_cumsum_VALUE (ca_size_t elements, ca_size_t min_count,
                       boolean8_t *m, void *ptr, CAStatIterator *it,
                       boolean8_t *retmask, float64_t *retval)
{
  volatile VALUE sum = rb_float_new(0.0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      if ( retmask ) {
        if ( count > min_count ) {
          *retmask = 1;
        }
        else {
          *retmask = 0;
        }
      }
      *retval = NUM2DBL(sum);
      retmask++; retval++;
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      *retval = NUM2DBL(sum);
      retval++;
      iterator_succ(it);
    }
  }
}

static ca_stat_proc_t
ca_proc_count[CA_NTYPE] = {
  NULL,
  ca_proc_count_boolean8_t,
  ca_proc_count_int8_t,
  ca_proc_count_uint8_t,
  ca_proc_count_int16_t,
  ca_proc_count_uint16_t,
  ca_proc_count_int32_t,
  ca_proc_count_uint32_t,
  ca_proc_count_int64_t,
  ca_proc_count_uint64_t,
  ca_proc_count_float32_t,
  ca_proc_count_float64_t,
  ca_proc_count_float128_t,
  NULL, /* ca_proc_count_cmplx64_t,  */
  NULL, /* ca_proc_count_cmplx128_t,    */
  NULL, /* ca_proc_count_cmplx256_t,      */
  ca_proc_count_VALUE, 
};

static ca_stat_proc_t
ca_proc_sum[CA_NTYPE] = {
  NULL,
  ca_proc_sum_boolean8_t,
  ca_proc_sum_int8_t,
  ca_proc_sum_uint8_t,
  ca_proc_sum_int16_t,
  ca_proc_sum_uint16_t,
  ca_proc_sum_int32_t,
  ca_proc_sum_uint32_t,
  ca_proc_sum_int64_t,
  ca_proc_sum_uint64_t,
  ca_proc_sum_float32_t,
  ca_proc_sum_float64_t,
  ca_proc_sum_float128_t,
  NULL, /* ca_proc_sum_cmplx64_t,  */
  NULL, /* ca_proc_sum_cmplx128_t,    */
  NULL, /* ca_proc_sum_cmplx256_t,      */
  ca_proc_sum_VALUE, 
};

static ca_stat_proc_t
ca_proc_mean[CA_NTYPE] = {
  NULL,
  ca_proc_mean_boolean8_t,
  ca_proc_mean_int8_t,
  ca_proc_mean_uint8_t,
  ca_proc_mean_int16_t,
  ca_proc_mean_uint16_t,
  ca_proc_mean_int32_t,
  ca_proc_mean_uint32_t,
  ca_proc_mean_int64_t,
  ca_proc_mean_uint64_t,
  ca_proc_mean_float32_t,
  ca_proc_mean_float64_t,
  ca_proc_mean_float128_t,
  NULL, /* ca_proc_mean_cmplx64_t,  */
  NULL, /* ca_proc_mean_cmplx128_t,    */
  NULL, /* ca_proc_mean_cmplx256_t,      */
  ca_proc_mean_VALUE, 
};

static ca_stat_proc_t
ca_proc_variancep[CA_NTYPE] = {
  NULL,
  ca_proc_variancep_boolean8_t,
  ca_proc_variancep_int8_t,
  ca_proc_variancep_uint8_t,
  ca_proc_variancep_int16_t,
  ca_proc_variancep_uint16_t,
  ca_proc_variancep_int32_t,
  ca_proc_variancep_uint32_t,
  ca_proc_variancep_int64_t,
  ca_proc_variancep_uint64_t,
  ca_proc_variancep_float32_t,
  ca_proc_variancep_float64_t,
  ca_proc_variancep_float128_t,
  NULL, /* ca_proc_variancep_cmplx64_t,  */
  NULL, /* ca_proc_variancep_cmplx128_t,    */
  NULL, /* ca_proc_variancep_cmplx256_t,      */
  ca_proc_variancep_VALUE, 
};

static ca_stat_proc_t
ca_proc_stddevp[CA_NTYPE] = {
  NULL,
  ca_proc_stddevp_boolean8_t,
  ca_proc_stddevp_int8_t,
  ca_proc_stddevp_uint8_t,
  ca_proc_stddevp_int16_t,
  ca_proc_stddevp_uint16_t,
  ca_proc_stddevp_int32_t,
  ca_proc_stddevp_uint32_t,
  ca_proc_stddevp_int64_t,
  ca_proc_stddevp_uint64_t,
  ca_proc_stddevp_float32_t,
  ca_proc_stddevp_float64_t,
  ca_proc_stddevp_float128_t,
  NULL, /* ca_proc_stddevp_cmplx64_t,  */
  NULL, /* ca_proc_stddevp_cmplx128_t,    */
  NULL, /* ca_proc_stddevp_cmplx256_t,      */
  ca_proc_stddevp_VALUE, 
};

static ca_stat_proc_t
ca_proc_variance[CA_NTYPE] = {
  NULL,
  ca_proc_variance_boolean8_t,
  ca_proc_variance_int8_t,
  ca_proc_variance_uint8_t,
  ca_proc_variance_int16_t,
  ca_proc_variance_uint16_t,
  ca_proc_variance_int32_t,
  ca_proc_variance_uint32_t,
  ca_proc_variance_int64_t,
  ca_proc_variance_uint64_t,
  ca_proc_variance_float32_t,
  ca_proc_variance_float64_t,
  ca_proc_variance_float128_t,
  NULL, /* ca_proc_variance_cmplx64_t,  */
  NULL, /* ca_proc_variance_cmplx128_t,    */
  NULL, /* ca_proc_variance_cmplx256_t,      */
  ca_proc_variance_VALUE, 
};

static ca_stat_proc_t
ca_proc_stddev[CA_NTYPE] = {
  NULL,
  ca_proc_stddev_boolean8_t,
  ca_proc_stddev_int8_t,
  ca_proc_stddev_uint8_t,
  ca_proc_stddev_int16_t,
  ca_proc_stddev_uint16_t,
  ca_proc_stddev_int32_t,
  ca_proc_stddev_uint32_t,
  ca_proc_stddev_int64_t,
  ca_proc_stddev_uint64_t,
  ca_proc_stddev_float32_t,
  ca_proc_stddev_float64_t,
  ca_proc_stddev_float128_t,
  NULL, /* ca_proc_stddev_cmplx64_t,  */
  NULL, /* ca_proc_stddev_cmplx128_t,    */
  NULL, /* ca_proc_stddev_cmplx256_t,      */
  ca_proc_stddev_VALUE, 
};

static ca_stat_proc_t
ca_proc_prod[CA_NTYPE] = {
  NULL,
  ca_proc_prod_boolean8_t,
  ca_proc_prod_int8_t,
  ca_proc_prod_uint8_t,
  ca_proc_prod_int16_t,
  ca_proc_prod_uint16_t,
  ca_proc_prod_int32_t,
  ca_proc_prod_uint32_t,
  ca_proc_prod_int64_t,
  ca_proc_prod_uint64_t,
  ca_proc_prod_float32_t,
  ca_proc_prod_float64_t,
  ca_proc_prod_float128_t,
  NULL, /* ca_proc_prod_cmplx64_t,  */
  NULL, /* ca_proc_prod_cmplx128_t,    */
  NULL, /* ca_proc_prod_cmplx256_t,      */
  ca_proc_prod_VALUE, 
};

static ca_stat_proc_t
ca_proc_min[CA_NTYPE] = {
  NULL,
  ca_proc_min_boolean8_t,
  ca_proc_min_int8_t,
  ca_proc_min_uint8_t,
  ca_proc_min_int16_t,
  ca_proc_min_uint16_t,
  ca_proc_min_int32_t,
  ca_proc_min_uint32_t,
  ca_proc_min_int64_t,
  ca_proc_min_uint64_t,
  ca_proc_min_float32_t,
  ca_proc_min_float64_t,
  ca_proc_min_float128_t,
  NULL, /* ca_proc_min_cmplx64_t,  */
  NULL, /* ca_proc_min_cmplx128_t,    */
  NULL, /* ca_proc_min_cmplx256_t,      */
  ca_proc_min_VALUE, 
};

static ca_stat_proc_t
ca_proc_min_addr[CA_NTYPE] = {
  NULL,
  ca_proc_min_addr_boolean8_t,
  ca_proc_min_addr_int8_t,
  ca_proc_min_addr_uint8_t,
  ca_proc_min_addr_int16_t,
  ca_proc_min_addr_uint16_t,
  ca_proc_min_addr_int32_t,
  ca_proc_min_addr_uint32_t,
  ca_proc_min_addr_int64_t,
  ca_proc_min_addr_uint64_t,
  ca_proc_min_addr_float32_t,
  ca_proc_min_addr_float64_t,
  ca_proc_min_addr_float128_t,
  NULL, /* ca_proc_min_addr_cmplx64_t,  */
  NULL, /* ca_proc_min_addr_cmplx128_t,    */
  NULL, /* ca_proc_min_addr_cmplx256_t,      */
  ca_proc_min_addr_VALUE, 
};

static ca_stat_proc_t
ca_proc_max[CA_NTYPE] = {
  NULL,
  ca_proc_max_boolean8_t,
  ca_proc_max_int8_t,
  ca_proc_max_uint8_t,
  ca_proc_max_int16_t,
  ca_proc_max_uint16_t,
  ca_proc_max_int32_t,
  ca_proc_max_uint32_t,
  ca_proc_max_int64_t,
  ca_proc_max_uint64_t,
  ca_proc_max_float32_t,
  ca_proc_max_float64_t,
  ca_proc_max_float128_t,
  NULL, /* ca_proc_max_cmplx64_t,  */
  NULL, /* ca_proc_max_cmplx128_t,    */
  NULL, /* ca_proc_max_cmplx256_t,      */
  ca_proc_max_VALUE, 
};

static ca_stat_proc_t
ca_proc_max_addr[CA_NTYPE] = {
  NULL,
  ca_proc_max_addr_boolean8_t,
  ca_proc_max_addr_int8_t,
  ca_proc_max_addr_uint8_t,
  ca_proc_max_addr_int16_t,
  ca_proc_max_addr_uint16_t,
  ca_proc_max_addr_int32_t,
  ca_proc_max_addr_uint32_t,
  ca_proc_max_addr_int64_t,
  ca_proc_max_addr_uint64_t,
  ca_proc_max_addr_float32_t,
  ca_proc_max_addr_float64_t,
  ca_proc_max_addr_float128_t,
  NULL, /* ca_proc_max_addr_cmplx64_t,  */
  NULL, /* ca_proc_max_addr_cmplx128_t,    */
  NULL, /* ca_proc_max_addr_cmplx256_t,      */
  ca_proc_max_addr_VALUE, 
};

static ca_stat_proc_t
ca_proc_cumcount[CA_NTYPE] = {
  NULL,
  ca_proc_cumcount_boolean8_t,
  ca_proc_cumcount_int8_t,
  ca_proc_cumcount_uint8_t,
  ca_proc_cumcount_int16_t,
  ca_proc_cumcount_uint16_t,
  ca_proc_cumcount_int32_t,
  ca_proc_cumcount_uint32_t,
  ca_proc_cumcount_int64_t,
  ca_proc_cumcount_uint64_t,
  ca_proc_cumcount_float32_t,
  ca_proc_cumcount_float64_t,
  ca_proc_cumcount_float128_t,
  NULL, /* ca_proc_cumcount_cmplx64_t,  */
  NULL, /* ca_proc_cumcount_cmplx128_t,    */
  NULL, /* ca_proc_cumcount_cmplx256_t,      */
  ca_proc_cumcount_VALUE, 
};

static ca_stat_proc_t
ca_proc_cumsum[CA_NTYPE] = {
  NULL,
  ca_proc_cumsum_boolean8_t,
  ca_proc_cumsum_int8_t,
  ca_proc_cumsum_uint8_t,
  ca_proc_cumsum_int16_t,
  ca_proc_cumsum_uint16_t,
  ca_proc_cumsum_int32_t,
  ca_proc_cumsum_uint32_t,
  ca_proc_cumsum_int64_t,
  ca_proc_cumsum_uint64_t,
  ca_proc_cumsum_float32_t,
  ca_proc_cumsum_float64_t,
  ca_proc_cumsum_float128_t,
  NULL, /* ca_proc_cumsum_cmplx64_t,  */
  NULL, /* ca_proc_cumsum_cmplx128_t,    */
  NULL, /* ca_proc_cumsum_cmplx256_t,      */
  ca_proc_cumsum_VALUE, 
};


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_boolean8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, boolean8_t *retval)
{
  boolean8_t sum = 0;
  boolean8_t *p = (boolean8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (boolean8_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (boolean8_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_int8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, int8_t *retval)
{
  int8_t sum = 0;
  int8_t *p = (int8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (int8_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (int8_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_uint8_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, uint8_t *retval)
{
  uint8_t sum = 0;
  uint8_t *p = (uint8_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (uint8_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (uint8_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_int16_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, int16_t *retval)
{
  int16_t sum = 0;
  int16_t *p = (int16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (int16_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (int16_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_uint16_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, uint16_t *retval)
{
  uint16_t sum = 0;
  uint16_t *p = (uint16_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (uint16_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (uint16_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_int32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, int32_t *retval)
{
  int32_t sum = 0;
  int32_t *p = (int32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (int32_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (int32_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_uint32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, uint32_t *retval)
{
  uint32_t sum = 0;
  uint32_t *p = (uint32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (uint32_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (uint32_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULONG2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_int64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, int64_t *retval)
{
  int64_t sum = 0;
  int64_t *p = (int64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (int64_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (int64_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : LL2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_uint64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, uint64_t *retval)
{
  uint64_t sum = 0;
  uint64_t *p = (uint64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (uint64_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (uint64_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : ULL2NUM(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_float32_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, float32_t *retval)
{
  float32_t sum = 0.0;
  float32_t *p = (float32_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (float32_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (float32_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_float64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, float64_t *retval)
{
  float64_t sum = 0.0;
  float64_t *p = (float64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (float64_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (float64_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_float128_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, float128_t *retval)
{
  float128_t sum = 0.0;
  float128_t *p = (float128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (float128_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (float128_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_float_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_cmplx64_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, cmplx64_t *retval)
{
  cmplx64_t sum = 0.0;
  cmplx64_t *p = (cmplx64_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (cmplx64_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (cmplx64_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_ccomplex_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_cmplx128_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, cmplx128_t *retval)
{
  cmplx128_t sum = 0.0;
  cmplx128_t *p = (cmplx128_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (cmplx128_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (cmplx128_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_ccomplex_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_cmplx256_t (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, cmplx256_t *retval)
{
  cmplx256_t sum = 0.0;
  cmplx256_t *p = (cmplx256_t *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_c(sum, (cmplx256_t)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_c(sum, (cmplx256_t)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : rb_ccomplex_new(sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}


/* ============================= */
/* ca_proc_accum                 */
/* ============================= */

static void
ca_proc_accum_VALUE (ca_size_t elements, ca_size_t min_count,
                      boolean8_t *m, void *ptr, CAStatIterator *it,
                      int return_object, VALUE *retobj,
                      boolean8_t *retmask, VALUE *retval)
{
  VALUE sum = INT2NUM(0);
  VALUE *p = (VALUE *) ptr;
  ca_size_t *a = (ca_size_t *) it;
  ca_size_t count = 0;
  ca_size_t i;
  iterator_rewind(it);
  if ( m ) {
    for (i=0; i<elements; i++) {
      if ( *(m + *a) ) {
        count++;
      }
      else {
        sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      }
      iterator_succ(it);
    }
  }
  else {
    for (i=0; i<elements; i++) {
      sum = add_VALUE(sum, (VALUE)(*(p + *a)));
      iterator_succ(it);
    }
  }
  if ( return_object ) {
    *retobj = ( count > min_count ) ? CA_UNDEF : (sum);
  }
  else {
    if ( retmask ) {
      *retmask = ( count > min_count ) ? 1 : 0;
    }
    *retval  = sum;
  }
}

static ca_stat_proc_t
ca_proc_accum[CA_NTYPE] = {
  NULL,
  ca_proc_accum_boolean8_t,
  ca_proc_accum_int8_t,
  ca_proc_accum_uint8_t,
  ca_proc_accum_int16_t,
  ca_proc_accum_uint16_t,
  ca_proc_accum_int32_t,
  ca_proc_accum_uint32_t,
  ca_proc_accum_int64_t,
  ca_proc_accum_uint64_t,
  ca_proc_accum_float32_t,
  ca_proc_accum_float64_t,
  ca_proc_accum_float128_t,
  ca_proc_accum_cmplx64_t,
  ca_proc_accum_cmplx128_t,
  ca_proc_accum_cmplx256_t,
  ca_proc_accum_VALUE, 
};


static VALUE
rb_ca_stat_1d (VALUE self, VALUE rmc, VALUE vfval,
               ca_stat_proc_t *ca_proc)
{
  volatile VALUE out;
  CArray *ca;
  CAStatIterator it;
  boolean8_t *m;
  ca_size_t mc;

  Data_Get_Struct(self, CArray, ca);

  if ( ! ca_proc[ca->data_type] ) {
    rb_raise(rb_eCADataTypeError,
             "this method is not implemented for data_type %s",
             ca_type_name[ca->data_type]);
  }

  if ( ca->elements == 0 ) {
    out = CA_UNDEF;
  }
  else {
    ca_attach(ca);
    m = ( ca->mask ) ? (boolean8_t *)ca->mask->ptr : NULL;
    mc = ( ( ! ca_has_mask(ca) ) || NIL_P(rmc)) ? ca->elements - 1 : NUM2SIZE(rmc);
    if ( mc < 0 ) {
      mc += ca->elements;
    }
    it.step = 0;
    ca_proc[ca->data_type](ca->elements, mc, m, ca->ptr, &it, 1, &out, NULL, NULL);
    ca_detach(ca);
  }

  if ( out == CA_UNDEF && ( vfval != CA_NIL ) ) {
    out = vfval;
  }

  return out;
}

static void
ca_stat_nd_contig_loop (CArray *ca, CArray *co, ca_size_t mc,
                        ca_stat_proc_t *ca_proc,
                        int level, ca_size_t *idx, char **op, boolean8_t **om)
{
  void *p;
  boolean8_t *m;
  ca_size_t i, n;
  if ( level == co->ndim ) {
    CAStatIterator it;
    n = ca->elements/co->elements;
    idx[level] = 0;
    p = ca_ptr_at_index(ca, idx);
    m = ( ca->mask && ca->mask->ptr ) ?
                          (boolean8_t *)ca_ptr_at_index(ca->mask, idx) : NULL;
		it.step = 0;
    ca_proc[ca->data_type](n, mc, m, p, &it, 0, NULL, *om, *op);
    if ( *om ) {
      *om += 1;
    }
    *op += co->bytes;
  }
  else {
    for (i=0; i<ca->dim[level]; i++) {
      idx[level] = i;
      ca_stat_nd_contig_loop(ca, co, mc, ca_proc, level+1, idx, op, om);
    }
  }
}

static VALUE
rb_ca_stat_nd_contig (VALUE self, VALUE vaxis, VALUE rmc, VALUE vfval,
                      int8_t data_type, ca_stat_proc_t *ca_proc)
{
  volatile VALUE out;
  CArray *ca, *co;
  ca_size_t mc, ndim;

  Data_Get_Struct(self, CArray, ca);

  if ( ! ca_proc[ca->data_type] ) {
    rb_raise(rb_eCADataTypeError,
             "this method is not implemented for data_type %s",
             ca_type_name[ca->data_type]);
  }

  ndim = ca->ndim - RARRAY_LEN(vaxis);
  if ( ndim <= 0 || ndim > ca->ndim ) {
    rb_raise(rb_eRuntimeError, "invalid dimension specified");
  }

  out = rb_carray_new(data_type, ndim, ca->dim, 0, NULL);
  Data_Get_Struct(out, CArray, co);

	if ( ca_has_mask(ca) ) {
		ca_create_mask(co);
	}

  if ( NIL_P(rmc) ) {
    mc = ca->elements/co->elements - 1;
  }
  else {
    mc = NUM2SIZE(rmc);
  }

  if ( mc < 0 ) {
    mc += ca->elements;
  }

  {
    ca_size_t idx[CA_RANK_MAX];
    char *op;
    boolean8_t *om;
    int8_t i;
    for (i=0; i<ca->ndim; i++) {
      idx[i] = 0;
    }
    ca_attach_n(2, ca, co);
    op = co->ptr;
    om = ( co->mask && co->mask->ptr ) ? (boolean8_t *)co->mask->ptr : NULL;
    ca_stat_nd_contig_loop(ca, co, mc, ca_proc, 0, idx, &op, &om);
    ca_sync(co);
    ca_detach_n(2, ca, co);
  }

  if ( ca_has_mask(co) && ( vfval != CA_NIL ) ) {
    out = rb_ca_mask_fill_copy(out, vfval);
  }

  return out;
}

static void
ca_stat_get_offset_loop (CArray *ca, ca_size_t *dm,
                        int level, ca_size_t *idx,
                        int level1, ca_size_t *idx1,
                        CArray *offset)
{
  ca_size_t i;
  if ( level == ca->ndim - 1 ) {
    if ( dm[level] == 0 ) {
      idx[level] = 0;
      *(ca_size_t *)ca_ptr_at_index(offset, idx1) = ca_index2addr(ca, idx);
    }
    else {
      for (i=0; i<ca->dim[level]; i++) {
        idx[level] = i;
        idx1[level1] = i;
        *(ca_size_t *)ca_ptr_at_index(offset, idx1) = ca_index2addr(ca, idx);
      }
    }
  }
  else {
    if ( dm[level] == 0 ) {
      idx[level] = 0;
      ca_stat_get_offset_loop(ca, dm, level+1, idx, level1, idx1, offset);
    }
    else {
      for (i=0; i<ca->dim[level]; i++) {
        idx[level] = i;
        idx1[level1] = i;
        ca_stat_get_offset_loop(ca, dm, level+1, idx, level1+1, idx1, offset);
      }
    }
  }
}

static VALUE
rb_ca_stat_nd_discrete (VALUE self, VALUE vaxis, VALUE rmc, VALUE vfval,
                        int8_t data_type, ca_stat_proc_t *ca_proc)
{
  volatile VALUE out;
  ca_size_t idx[CA_RANK_MAX];
  ca_size_t idx1[CA_RANK_MAX];
  ca_size_t out_dim[CA_RANK_MAX];
  ca_size_t loop_dim[CA_RANK_MAX];
  ca_size_t dm[CA_RANK_MAX], dn[CA_RANK_MAX];
  CArray *ca, *co, *first, *offset;
  ca_size_t out_ndim, loop_ndim;
  ca_size_t mc;
  ca_size_t i, k;

  Data_Get_Struct(self, CArray, ca);

  for (i=0; i<ca->ndim; i++) {
    dm[i] = 0;
    dn[i] = 1;
  }

  for (i=0; i<RARRAY_LEN(vaxis); i++) {
    k = NUM2SIZE(rb_ary_entry(vaxis, i));
    dm[k] = 1;
    dn[k] = 0;
  }

  out_ndim  = 0;
  loop_ndim = 0;
  for (i=0; i<ca->ndim; i++) {
    if ( dm[i] ) {
      loop_dim[loop_ndim] = ca->dim[i];
      loop_ndim += 1;
    }
    else {
      out_dim[out_ndim] = ca->dim[i];
      out_ndim += 1;
    }
  }

  out    = rb_carray_new(data_type, out_ndim, out_dim, 0, NULL);
  Data_Get_Struct(out, CArray, co);
  
  first  = carray_new(CA_SIZE, out_ndim, out_dim, 0, NULL);
  first->ptr = realloc(first->ptr, first->bytes*(first->elements+1));

  offset = carray_new(CA_SIZE, loop_ndim, loop_dim, 0, NULL);
  offset->ptr = realloc(offset->ptr, offset->bytes*(offset->elements+1));

  ca_stat_get_offset_loop(ca, dn, 0, idx, 0, idx1, first);
  ca_stat_get_offset_loop(ca, dm, 0, idx, 0, idx1, offset);

	if ( ca_has_mask(ca) ) {
		ca_create_mask(co);
	}

  if ( NIL_P(rmc) ) {
    mc = offset->elements - 1;
  }
  else {
    mc = NUM2SIZE(rmc);
  }

  if ( mc < 0 ) {
    mc += offset->elements;
  }

  ca_attach_n(2, ca, co);

  {
    CAStatIterator it, jt;
    ca_size_t *a = (ca_size_t *) (&it);
    boolean8_t *m0, *m, *om;
    char *p, *op;

    it.step = 1;
    it.addr = (ca_size_t *)first->ptr;
    jt.step = 1;
    jt.addr = (ca_size_t *)offset->ptr;

    m0 = ( ca->mask && ca->mask->ptr ) ? (boolean8_t *)ca->mask->ptr : NULL;
    om = ( co->mask && co->mask->ptr ) ? (boolean8_t *)co->mask->ptr : NULL;
    op = co->ptr;

    iterator_rewind(&it);
    for (i=0; i<co->elements; i++) {
      m = ( m0 ) ? m0 + (*a) : NULL;
      p = ca->ptr + (*a) * ca->bytes;
      ca_proc[ca->data_type](offset->elements, mc, m, p, &jt, 0, NULL, om, op);
      if ( om ) {
        om += 1;
      }
      op += co->bytes;
      iterator_succ(&it);
    }
  }

  ca_sync(co);
  ca_detach_n(2, ca, co);

  if ( ca_has_mask(co) && ( vfval != CA_NIL ) ) {
    out = rb_ca_mask_fill_copy(out, vfval);
  }

  ca_free(first);
  ca_free(offset);

  return out;
}

static VALUE
rb_ca_stat_general (int argc, VALUE *argv, VALUE self,
                    int8_t data_type, ca_stat_proc_t *ca_proc)
{
  volatile VALUE ropt, rmc = Qnil, rmask_limit = Qnil, rmin_count = Qnil, 
                 vfval = CA_NIL, vaxis;
  CArray *ca;

  Data_Get_Struct(self, CArray, ca);

  ropt = rb_pop_options(&argc, &argv);
  rb_scan_options(ropt,
                  "mask_limit,fill_value,min_count",
                  &rmask_limit, &vfval, &rmin_count);

  if ( ( ! NIL_P(rmask_limit) ) && ( ! NIL_P(rmin_count) ) ) {
    rb_raise(rb_eArgError,
             "don't specify mask_limit and min_count simaltaniously");
  }
  else if ( ! NIL_P(rmin_count) ) {
    ca_size_t min_count = NUM2SIZE(rmin_count);
    if ( min_count == 0 ) {
      rmc = Qnil;
    }
    else {
      rmc = SIZE2NUM(-min_count);
    }
  }
  else if ( ! NIL_P(rmask_limit) ) {
    ca_size_t mask_limit = NUM2SIZE(rmask_limit);
    if ( mask_limit == 0 ) {
      rmc = Qnil;
    }
    else {
      rmc = SIZE2NUM(mask_limit-1);
    }
  }
  else {
    rmc = Qnil;
  }

  if ( argc > 0 ) {
    vaxis = rb_ary_new4(argc, argv);
  }
  else {
    vaxis = Qnil;
  }

  if ( NIL_P(vaxis) ) {
    return rb_ca_stat_1d(self, rmc, vfval, ca_proc);
  }
  else {
    int is_contig = 1;
    ca_size_t i, k;

    vaxis = rb_funcall(vaxis, rb_intern("sort"), 0);
    vaxis = rb_funcall(vaxis, rb_intern("uniq"), 0);

    for (i=0; i<RARRAY_LEN(vaxis); i++) {
      k = NUM2SIZE(rb_ary_entry(vaxis, RARRAY_LEN(vaxis)-1-i));
      CA_CHECK_INDEX(k, ca->ndim);
      if ( k != ca->ndim-1-i ) {
        is_contig = 0;
      }
    }

    if ( is_contig ) {
      if ( RARRAY_LEN(vaxis) == ca->ndim ) {
        return rb_ca_stat_1d(self, rmc, vfval, ca_proc);
      }
      else {
        return rb_ca_stat_nd_contig(self, vaxis, rmc, vfval,
                                    data_type, ca_proc);
      }
    }
    else {
      return rb_ca_stat_nd_discrete(self, vaxis, rmc, vfval,
                                    data_type, ca_proc);
    }
  }
}


static VALUE
rb_ca_prod (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_prod);
}

static VALUE
rb_ca_count (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_INT32, ca_proc_count);
}

static VALUE
rb_ca_sum (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_sum);
}

static VALUE
rb_ca_mean (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_mean);
}

static VALUE
rb_ca_variancep (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_variancep);
}

static VALUE
rb_ca_stddevp (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_stddevp);
}

static VALUE
rb_ca_variance (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_variance);
}

static VALUE
rb_ca_stddev (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_FLOAT64, ca_proc_stddev);
}

static VALUE
rb_ca_min (int argc, VALUE *argv, VALUE self)
{
  CArray *ca;
  Data_Get_Struct(self, CArray, ca);
  return rb_ca_stat_general(argc, argv, self, ca->data_type, ca_proc_min);
}

static VALUE
rb_ca_min_addr (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_INT32, ca_proc_min_addr);
}

static VALUE
rb_ca_max (int argc, VALUE *argv, VALUE self)
{
  CArray *ca;
  Data_Get_Struct(self, CArray, ca);
  return rb_ca_stat_general(argc, argv, self, ca->data_type, ca_proc_max);
}

static VALUE
rb_ca_max_addr (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_general(argc, argv, self, CA_INT32, ca_proc_max_addr);
}

static VALUE
rb_ca_accum (int argc, VALUE *argv, VALUE self)
{
  CArray *ca;
  Data_Get_Struct(self, CArray, ca);
  return rb_ca_stat_general(argc, argv, self, ca->data_type, ca_proc_accum);
}

static VALUE
rb_ca_stat_type2 (int argc, VALUE *argv, VALUE self,
                  int8_t data_type, ca_stat_proc_t *ca_proc)
{
  volatile VALUE out, rmc, rfval;
  CArray *ca, *co;
  ca_size_t mc;

  rb_scan_args(argc, argv, "02", (VALUE *) &rmc, (VALUE *) &rfval);

  Data_Get_Struct(self, CArray, ca);

  if ( ! ca_proc[ca->data_type] ) {
    rb_raise(rb_eCADataTypeError,
             "this method is not implemented for data_type %s",
             ca_type_name[ca->data_type]);
  }

  out = rb_carray_new(data_type, 1, &ca->elements, 0, NULL);
  Data_Get_Struct(out, CArray, co);

  if ( ca->elements == 0 ) {
    ;
  }
  else {
    CAStatIterator it;
    boolean8_t *m = NULL;
    boolean8_t *om = NULL;
    ca_attach(ca);
    if ( ca->mask ) {
      m = (boolean8_t *)ca->mask->ptr;
    }
    mc = ( (! ca->mask) || NIL_P(rmc) ) ? ca->elements - 1 : NUM2SIZE(rmc);
    if ( mc < 0 ) {
      mc += ca->elements;
    }
    if ( m ) {
      ca_create_mask(co);
      om = (boolean8_t *) co->mask->ptr;
    }
    it.step = 0;
    ca_proc[ca->data_type](ca->elements, mc, m, ca->ptr, &it,
                           om, co->ptr);
    ca_detach(ca);
  }

  if ( ca_has_mask(co) && ( ! NIL_P(rfval) ) ) {
    out = rb_ca_mask_fill_copy(out, rfval);
  }

  return out;
}

static void
ca_dimstat_type2_loop (CArray *ca, CArray *co, ca_size_t mc,
                       ca_stat_proc_t *ca_proc,
                       int level, ca_size_t *idx, char **op, boolean8_t **om)
{
  void *p;
  boolean8_t *m;
  ca_size_t i, n;
  if ( level == co->ndim ) {
    CAStatIterator it;
    n = ca->elements/co->elements;
    idx[level] = 0;
    p = ca_ptr_at_index(ca, idx);
    m = ( ca->mask && ca->mask->ptr ) ?
                                (boolean8_t *)ca_ptr_at_index(ca->mask, idx) : NULL;
    it.step = 0;
    ca_proc[ca->data_type](n, mc, m, p, &it, 0, NULL, *om, *op);
    printf("world\n");
    if ( *om ) {
      *om += n;
    }
    *op += n*co->bytes;
  }
  else {
    for (i=0; i<ca->dim[level]; i++) {
      idx[level] = i;
      ca_dimstat_type2_loop(ca, co, mc, ca_proc, level+1, idx, op, om);
    }
  }
}

static void
ca_dimstat_type2 (CArray *ca, CArray *co, ca_size_t mc, ca_stat_proc_t *ca_proc)
{
  ca_size_t idx[CA_RANK_MAX];
  char *op;
  boolean8_t *om;
  int8_t i;
  for (i=0; i<ca->ndim; i++) {
    idx[i] = 0;
  }
  ca_attach_n(2, ca, co);
  op = co->ptr;
  om = ( co->mask && co->mask->ptr ) ? (boolean8_t *)co->mask->ptr : NULL;
  ca_dimstat_type2_loop(ca, co, mc, ca_proc, 0, idx, &op, &om);
  ca_sync(co);
  ca_detach_n(2, ca, co);
}

static VALUE
rb_ca_dimstat_type2 (int argc, VALUE *argv, VALUE self,
                    int8_t data_type, ca_stat_proc_t *ca_proc)
{
  volatile VALUE rndim, rmc, rfval;
  volatile VALUE out;
  CArray *ca, *co;
  ca_size_t odim[CA_RANK_MAX], n;
  ca_size_t mc, ndim;
  int32_t i;

  rb_scan_args(argc, argv, "12", (VALUE *) &rndim, (VALUE *) &rmc, (VALUE *) &rfval);

  Data_Get_Struct(self, CArray, ca);

  if ( ! ca_proc[ca->data_type] ) {
    rb_raise(rb_eCADataTypeError,
             "this method is not implemented for data_type %s",
             ca_type_name[ca->data_type]);
  }

  ndim = NUM2SIZE(rndim);
  if ( ndim <= 0 || ndim > ca->ndim ) {
    rb_raise(rb_eRuntimeError, "invalid dimension specified");
  }

  n = 1;
  for (i=0; i<ndim; i++) {
    odim[i] = ca->dim[i];
    n *= odim[i];
  }
  odim[ndim] = ca->elements/n;

  out = rb_carray_new(data_type, ndim+1, odim, 0, NULL);
  Data_Get_Struct(out, CArray, co);

	if ( ca_has_mask(ca) ) {
		ca_create_mask(co);
	}

  if ( NIL_P(rmc) ) {
    mc = ca->elements/co->elements - 1;
  }
  else {
    mc = NUM2SIZE(rmc);
  }

  if ( mc < 0 ) {
    mc += ca->elements;
  }

  ca_dimstat_type2(ca, co, mc, ca_proc);

  if ( ca_has_mask(co) && ( ! NIL_P(rfval) ) ) {
    out = rb_ca_mask_fill_copy(out, rfval);
  }

  return out;
}

static VALUE
rb_ca_cumcount (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_type2(argc, argv, self, CA_INT32, ca_proc_cumcount);
}

static VALUE
rb_ca_dimcumcount (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_dimstat_type2(argc, argv, self, CA_INT32, ca_proc_cumcount);
}

static VALUE
rb_ca_cumsum (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_stat_type2(argc, argv, self, CA_FLOAT64, ca_proc_cumsum);
}

static VALUE
rb_ca_dimcumsum (int argc, VALUE *argv, VALUE self)
{
  return rb_ca_dimstat_type2(argc, argv, self, CA_FLOAT64, ca_proc_cumsum);
}

void
Init_carray_stat_proc ()
{
  id_lt    = rb_intern("<");
  id_gt    = rb_intern(">");  
  id_plus  = rb_intern("+");
  id_minus = rb_intern("-"); 
  id_star  = rb_intern("*"); 
  id_quo   = rb_intern("quo");   

  rb_define_method(rb_cCArray, "prod",    rb_ca_prod, -1);
  rb_define_method(rb_cCArray, "count_valid", rb_ca_count, -1);
  rb_define_method(rb_cCArray, "sum",     rb_ca_sum, -1);
  rb_define_method(rb_cCArray, "mean",    rb_ca_mean, -1);
  rb_define_method(rb_cCArray, "variancep", rb_ca_variancep, -1);
  rb_define_method(rb_cCArray, "stddevp", rb_ca_stddevp, -1);
  rb_define_method(rb_cCArray, "variance",  rb_ca_variance, -1);
  rb_define_method(rb_cCArray, "stddev",  rb_ca_stddev, -1);
  rb_define_method(rb_cCArray, "min",     rb_ca_min, -1);
  rb_define_method(rb_cCArray, "max",     rb_ca_max, -1);
  rb_define_method(rb_cCArray, "min_addr", rb_ca_min_addr, -1);
  rb_define_method(rb_cCArray, "max_addr", rb_ca_max_addr, -1);

  rb_define_method(rb_cCArray, "accumulate",  rb_ca_accum, -1);

  rb_define_method(rb_cCArray, "cumcount",    rb_ca_cumcount, -1);
  rb_define_method(rb_cCArray, "dimcumcount", rb_ca_dimcumcount, -1);
  rb_define_method(rb_cCArray, "cumsum",    rb_ca_cumsum, -1);
  rb_define_method(rb_cCArray, "dimcumsum", rb_ca_dimcumsum, -1);
}



